from sqlalchemy.orm import aliased
from sqlalchemy import func, or_, and_

from zou.app import config
from zou.app.models.comment import Comment
from zou.app.models.entity import Entity
from zou.app.models.entity_type import EntityType
from zou.app.models.notification import Notification
from zou.app.models.person import Person
from zou.app.models.project import Project, ProjectPersonLink
from zou.app.models.project_status import ProjectStatus
from zou.app.models.subscription import Subscription
from zou.app.models.search_filter import SearchFilter
from zou.app.models.search_filter_group import SearchFilterGroup
from zou.app.models.task import Task
from zou.app.models.task_type import TaskType


from zou.app.services import (
    assets_service,
    custom_actions_service,
    edits_service,
    entities_service,
    notifications_service,
    names_service,
    persons_service,
    projects_service,
    shots_service,
    status_automations_service,
    tasks_service,
    files_service,
)
from zou.app.services.exception import (
    SearchFilterNotFoundException,
    SearchFilterGroupNotFoundException,
    NotificationNotFoundException,
    WrongParameterException,
)
from zou.app.utils import cache, fields, permissions, events


def clear_filter_cache(user_id=None):
    if user_id is None:
        cache.cache.delete_memoized(get_user_filters)
    else:
        cache.cache.delete_memoized(get_user_filters, user_id)


def clear_filter_group_cache(user_id=None):
    if user_id is None:
        cache.cache.delete_memoized(get_user_filter_groups)
    else:
        cache.cache.delete_memoized(get_user_filter_groups, user_id)


def clear_project_cache():
    cache.cache.delete_memoized(get_open_projects)


def build_assignee_filter():
    """
    Query filter for task to retrieve only tasks assigned to current user.
    """
    current_user = persons_service.get_current_user_raw()
    return Task.assignees.contains(current_user)


def build_team_filter():
    """
    Query filter for task to retrieve only models from project for which the
    user is part of the team.
    """
    current_user = persons_service.get_current_user_raw()
    return Project.team.contains(current_user)


def build_open_project_filter():
    """
    Query filter for project to retrieve only open projects.
    """
    open_status = projects_service.get_open_status()
    return Project.project_status_id == open_status["id"]


def build_related_projects_filter():
    """
    Query filter for project to retrieve open projects of which the user
    is part of the team.
    """
    projects = related_projects()
    project_ids = [project["id"] for project in projects]
    if len(project_ids) > 0:
        return Project.id.in_(project_ids)
    else:
        return Project.id.in_(["00000000-0000-0000-0000-000000000000"])


def related_projects():
    """
    Return all projects related to current user: open projects of which the user
    is part of the team as dicts.
    """
    persons_service.get_current_user()
    projects = related_projects_raw()
    return Project.serialize_list(projects)


def related_projects_raw():
    """
    Return all projects related to current user: open projects of which the user
    is part of the team as models.
    """
    current_user = persons_service.get_current_user()
    projects = (
        Project.query.join(
            ProjectStatus, Project.project_status_id == ProjectStatus.id
        )
        .join(ProjectPersonLink, Project.id == ProjectPersonLink.project_id)
        .filter(ProjectPersonLink.person_id == current_user["id"])
        .filter(build_open_project_filter())
        .distinct()
        .all()
    )
    return projects


def get_todos():
    """
    Get all unfinished tasks assigned to current user.
    """
    current_user = persons_service.get_current_user()
    projects = related_projects()
    return tasks_service.get_person_tasks(current_user["id"], projects)


def get_done_tasks():
    """
    Get all finished tasks assigned to current user for open projects.
    """
    current_user = persons_service.get_current_user()
    projects = related_projects()
    return tasks_service.get_person_done_tasks(current_user["id"], projects)


def get_tasks_to_check():
    """
    Get all tasks waiting for feedback in the user department.
    """
    departments_ids = None
    if permissions.has_admin_permissions():
        project_ids = None
    elif permissions.has_manager_permissions():
        project_ids = [project["id"] for project in related_projects()]
    elif permissions.has_supervisor_permissions():
        current_user = persons_service.get_current_user(relations=True)
        departments_ids = current_user["departments"]
        project_ids = [project["id"] for project in related_projects()]
    else:
        return []

    return tasks_service.get_person_tasks_to_check(
        project_ids, departments_ids
    )


def get_tasks_for_entity(entity_id):
    """
    Get all tasks assigned to current user and related to given entity.
    """
    query = (
        Task.query.join(Project)
        .join(ProjectStatus, Project.project_status_id == ProjectStatus.id)
        .filter(Task.entity_id == entity_id)
        .filter(build_assignee_filter())
        .filter(build_open_project_filter())
    )

    return fields.serialize_value(query.all())


def get_task_types_for_entity(entity_id):
    """
    Get all task types of tasks assigned to current user and related to given
    entity.
    """
    query = (
        TaskType.query.join(Task)
        .join(Project)
        .join(ProjectStatus, Project.project_status_id == ProjectStatus.id)
        .filter(Task.entity_id == entity_id)
        .filter(build_assignee_filter())
        .filter(build_open_project_filter())
    )

    return fields.serialize_value(query.all())


def get_assets_for_asset_type(project_id, asset_type_id):
    """
    Get all assets for given asset type anp project and for which user has
    a task related.
    """
    query = (
        Entity.query.join(EntityType)
        .join(Project)
        .join(Task, Task.entity_id == Entity.id)
        .join(ProjectStatus, Project.project_status_id == ProjectStatus.id)
        .filter(EntityType.id == asset_type_id)
        .filter(Project.id == project_id)
        .filter(build_assignee_filter())
        .filter(build_open_project_filter())
    )

    return Entity.serialize_list(query.all(), obj_type="Asset")


def get_asset_types_for_project(project_id):
    """
    Get all asset types for which there is an asset for which current user has a
    task assigned. Assets are listed in given project.
    """
    query = (
        EntityType.query.join(Entity, Entity.entity_type_id == EntityType.id)
        .join(Task, Task.entity_id == Entity.id)
        .join(Project)
        .join(ProjectStatus, Project.project_status_id == ProjectStatus.id)
        .filter(Project.id == project_id)
        .filter(build_assignee_filter())
        .filter(build_open_project_filter())
        .filter(assets_service.build_asset_type_filter())
    )

    return EntityType.serialize_list(query.all(), obj_type="AssetType")


def get_sequences_for_project(project_id):
    """
    Return all sequences for given project and for which current user has
    a task assigned to a shot.
    """
    shot_type = shots_service.get_shot_type()
    sequence_type = shots_service.get_sequence_type()

    Shot = aliased(Entity, name="shot")
    query = (
        Entity.query.join(Shot, Shot.parent_id == Entity.id)
        .join(Task, Task.entity_id == Shot.id)
        .join(EntityType, EntityType.id == Entity.entity_type_id)
        .join(Project, Project.id == Entity.project_id)
        .join(ProjectStatus, Project.project_status_id == ProjectStatus.id)
        .filter(Shot.entity_type_id == shot_type["id"])
        .filter(Entity.entity_type_id == sequence_type["id"])
        .filter(Project.id == project_id)
        .filter(build_assignee_filter())
        .filter(build_open_project_filter())
    )

    return Entity.serialize_list(query.all(), obj_type="Sequence")


def get_project_episodes(project_id):
    """
    Return all episodes for given project and for which current user has
    a task assigned to a shot.
    """
    shot_type = shots_service.get_shot_type()
    sequence_type = shots_service.get_sequence_type()
    episode_type = shots_service.get_episode_type()

    Shot = aliased(Entity, name="shot")
    Sequence = aliased(Entity, name="sequence")
    query = (
        Entity.query.join(Sequence, Sequence.parent_id == Entity.id)
        .join(Shot, Shot.parent_id == Sequence.id)
        .join(Task, Task.entity_id == Shot.id)
        .join(Project, Project.id == Entity.project_id)
        .join(ProjectStatus, Project.project_status_id == ProjectStatus.id)
        .filter(Shot.entity_type_id == shot_type["id"])
        .filter(Sequence.entity_type_id == sequence_type["id"])
        .filter(Entity.entity_type_id == episode_type["id"])
        .filter(Project.id == project_id)
        .filter(build_assignee_filter())
        .filter(build_open_project_filter())
    )

    return Entity.serialize_list(query.all(), obj_type="Episode")


def get_shots_for_sequence(sequence_id):
    """
    Get all shots for given sequence and for which the user has a task assigned.
    """
    shot_type = shots_service.get_shot_type()
    query = (
        Entity.query.join(Task)
        .join(Project)
        .join(ProjectStatus, Project.project_status_id == ProjectStatus.id)
        .join(EntityType)
        .filter(Entity.entity_type_id == shot_type["id"])
        .filter(Entity.parent_id == sequence_id)
        .filter(build_assignee_filter())
        .filter(build_open_project_filter())
    )

    return Entity.serialize_list(query.all(), obj_type="Shot")


def get_scenes_for_sequence(sequence_id):
    """
    Get all layout scenes for given sequence and for which the user has a task
    assigned.
    """
    scene_type = shots_service.get_scene_type()
    query = (
        Entity.query.join(Task)
        .join(Project)
        .join(ProjectStatus, Project.project_status_id == ProjectStatus.id)
        .join(EntityType)
        .filter(Entity.entity_type_id == scene_type["id"])
        .filter(Entity.parent_id == sequence_id)
        .filter(build_assignee_filter())
        .filter(build_open_project_filter())
    )

    return Entity.serialize_list(query.all(), obj_type="Scene")


def get_open_projects(name=None):
    """
    Get all open projects for which current user is part of the team.
    """
    query = Project.query.join(
        ProjectStatus, Project.project_status_id == ProjectStatus.id
    ).filter(build_open_project_filter())

    if name is not None:
        query = query.filter(Project.name == name)

    if not permissions.has_admin_permissions():
        current_user = persons_service.get_current_user()
        query = query.join(
            ProjectPersonLink, Project.id == ProjectPersonLink.project_id
        )
        query = query.filter(ProjectPersonLink.person_id == current_user["id"])

    for_client = False
    vendor_departments = None
    if permissions.has_client_permissions():
        for_client = True
    elif permissions.has_vendor_permissions():
        vendor_departments = persons_service.get_current_user(relations=True)[
            "departments"
        ]

    return projects_service.get_projects_with_extra_data(
        query, for_client, vendor_departments
    )


def get_open_project_ids():
    """
    Get all open project ids for which current user is part of the team.
    """
    return [project["id"] for project in get_open_projects()]


def get_projects(name=None):
    """
    Get all projects for which current user is part of the team.
    """
    current_user = persons_service.get_current_user()
    query = (
        Project.query.join(
            ProjectStatus, Project.project_status_id == ProjectStatus.id
        )
        .join(ProjectPersonLink, Project.id == ProjectPersonLink.project_id)
        .filter(ProjectPersonLink.person_id == current_user["id"])
    )

    if name is not None:
        query = query.filter(Project.name == name)

    return fields.serialize_value(query.all())


def check_working_on_entity(entity_id):
    """
    Return True if user has task assigned which is related to given entity.
    """
    current_user = persons_service.get_current_user_raw()
    query = Task.query.filter(Task.assignees.contains(current_user)).filter(
        Task.entity_id == entity_id
    )

    if query.first() is None:
        raise permissions.PermissionDenied

    return True


def check_working_on_task(task_id):
    """
    Return True if user has task assigned.
    """
    current_user = persons_service.get_current_user_raw()
    query = Task.query.filter(Task.assignees.contains(current_user)).filter(
        Task.id == task_id
    )

    if query.first() is None:
        raise permissions.PermissionDenied

    return True


def check_person_access(person_id):
    """
    Return True if user is an admin or is matching given person id.
    """
    if (
        permissions.has_admin_permissions()
        or persons_service.get_current_user()["id"] == person_id
    ):
        return True
    else:
        raise permissions.PermissionDenied


def check_belong_to_project(project_id):
    """
    Return true if current user is assigned to a task of the given project or
    if current_user is part of the project team.
    """
    if project_id is None:
        return False

    project = projects_service.get_project(str(project_id), relations=True)
    current_user = persons_service.get_current_user()
    return current_user["id"] in project["team"]


def has_project_access(project_id):
    """
    Return true if current user is an admin or has a task assigned for this
    project.
    """
    return permissions.has_admin_permissions() or check_belong_to_project(
        project_id
    )


def check_project_access(project_id):
    """
    Return true if current user is a manager or has a task assigned for this
    project. Raise a PermissionDenied exception if not.
    """
    is_allowed = has_project_access(project_id)
    if not is_allowed:
        raise permissions.PermissionDenied
    return is_allowed


def block_access_to_vendor():
    """
    Raise PermissionDenied if current user has a vendor role.
    """
    if permissions.has_vendor_permissions():
        raise permissions.PermissionDenied
    return True


def check_entity_access(entity_id):
    """
    Return true if current user is not a vendor or has a task assigned for this
    project.
    """
    is_allowed = not permissions.has_vendor_permissions()
    if not is_allowed:
        nb_tasks = (
            Task.query.filter(Task.entity_id == entity_id)
            .filter(build_assignee_filter())
            .count()
        )
        if nb_tasks == 0:
            raise permissions.PermissionDenied
        is_allowed = True
    return is_allowed


def check_task_status_access(task_status_id):
    """
    Return true if current user can use this task status.
    """
    is_artist = permissions.has_artist_permissions()
    is_client = permissions.has_client_permissions()
    if is_artist or is_client:
        task_status = tasks_service.get_task_status(task_status_id)
        if is_artist and not task_status["is_artist_allowed"]:
            raise permissions.PermissionDenied
        if is_client and not task_status["is_client_allowed"]:
            raise permissions.PermissionDenied
    return True


def check_task_access(task_id):
    """
    Return true if current user can have access to a task.
    """
    task = tasks_service.get_task(task_id)
    check_project_access(task["project_id"])
    check_entity_access(task["entity_id"])
    return True


def check_task_action_access(task_id):
    """
    Return true if current user can have access to a task action.
    """
    task = tasks_service.get_task(task_id, relations=True)
    is_allowed = False
    if permissions.has_admin_permissions():
        is_allowed = True
    elif check_belong_to_project(task["project_id"]):
        if (
            permissions.has_manager_permissions()
            or permissions.has_client_permissions()
        ):
            is_allowed = True
        else:
            user = persons_service.get_current_user(relations=True)
            is_allowed = user["id"] in task["assignees"]
            if not is_allowed and permissions.has_supervisor_permissions():
                is_allowed = (
                    user["departments"] == []
                    or tasks_service.get_task_type(task["task_type_id"])[
                        "department_id"
                    ]
                    in user["departments"]
                )

    if not is_allowed:
        raise permissions.PermissionDenied
    return is_allowed


def check_supervisor_project_task_type_access(project_id, task_type_id):
    """
    Return true if current user can have access to a task type.
    """
    is_allowed = False
    if permissions.has_admin_permissions() or (
        permissions.has_manager_permissions()
        and check_belong_to_project(project_id)
    ):
        is_allowed = True
    elif permissions.has_supervisor_permissions() and check_belong_to_project(
        project_id
    ):
        user = persons_service.get_current_user(relations=True)
        is_allowed = (
            user["departments"] == []
            or tasks_service.get_task_type(task_type_id)["department_id"]
            in user["departments"]
        )

    if not is_allowed:
        raise permissions.PermissionDenied
    return is_allowed


def check_comment_access(comment_id):
    """
    Return true if current user can have access to a comment.
    """
    if permissions.has_admin_permissions():
        return True
    else:
        comment = tasks_service.get_comment(comment_id)
        person_id = comment["person_id"]
        task_id = comment["object_id"]
        task = tasks_service.get_task(task_id)
        if task is None:
            tasks_service.clear_task_cache(task_id)
            task = tasks_service.get_task(task_id)
        check_project_access(task["project_id"])
        check_entity_access(task["entity_id"])

        if (
            permissions.has_supervisor_permissions()
            or permissions.has_manager_permissions()
        ):
            return True
        elif permissions.has_client_permissions():
            current_user = persons_service.get_current_user()
            project = projects_service.get_project(task["project_id"])
            if project.get("is_clients_isolated", False):
                if not comment["person_id"] == current_user["id"]:
                    raise permissions.PermissionDenied
            if persons_service.get_person(person_id)["role"] == "client":
                return True
            else:
                raise permissions.PermissionDenied
        elif persons_service.get_person(person_id)["role"] == "client":
            raise permissions.PermissionDenied

        return True


def has_manager_project_access(project_id):
    """
    Return true if current user is a manager and has a task assigned for this
    project.
    """
    return permissions.has_admin_permissions() or (
        permissions.has_manager_permissions()
        and check_belong_to_project(project_id)
    )


def check_manager_project_access(project_id):
    """
    Return true if current user is a manager and has a task assigned for this
    project. Raise a PermissionDenied exception if not.
    """
    is_allowed = has_manager_project_access(project_id)
    if not is_allowed:
        raise permissions.PermissionDenied
    return is_allowed


def check_time_spent_access(task_id, person_id):
    """
    Return true if current user is an admin or is a manager or is assigned to
    the task.
    """
    task = tasks_service.get_task(task_id, relations=True)
    is_allowed = person_id in task["assignees"] and (
        persons_service.get_current_user()["id"] == person_id
        or (
            permissions.has_admin_permissions()
            or (
                permissions.has_manager_permissions()
                and check_belong_to_project(task["project_id"])
            )
        )
    )

    if not is_allowed:
        raise permissions.PermissionDenied
    return is_allowed


def check_supervisor_project_access(project_id):
    """
    Return true if current user is a manager or a supervisor and has a task
    assigned for this project.
    """
    is_allowed = permissions.has_admin_permissions() or (
        (
            permissions.has_manager_permissions()
            or permissions.has_supervisor_permissions()
        )
        and check_belong_to_project(project_id)
    )
    if not is_allowed:
        raise permissions.PermissionDenied
    return is_allowed


def check_supervisor_task_access(task, new_data={}):
    """
    Return true if current user is a manager and has a task assigned related
    to the project of this task or is a supervisor and can modify data accorded
    to his departments
    """
    is_allowed = False
    if permissions.has_admin_permissions() or (
        permissions.has_manager_permissions()
        and check_belong_to_project(task["project_id"])
    ):
        is_allowed = True
    elif permissions.has_supervisor_permissions() and check_belong_to_project(
        task["project_id"]
    ):
        # checks that the supervisor only modifies columns
        # for which he is authorized
        allowed_columns = set(
            ["priority", "start_date", "due_date", "estimation", "difficulty"]
        )
        if len(set(new_data.keys()) - allowed_columns) == 0:
            user_departments = persons_service.get_current_user(
                relations=True
            )["departments"]
            if (
                user_departments == []
                or tasks_service.get_task_type(task["task_type_id"])[
                    "department_id"
                ]
                in user_departments
            ):
                is_allowed = True

    if not is_allowed:
        raise permissions.PermissionDenied
    return is_allowed


def check_metadata_department_access(entity, new_data={}):
    """
    Return true if current user is a manager and has a task assigned for this
    project or is a supervisor and is allowed to modify data accorded to
    his departments
    """
    is_allowed = False
    if permissions.has_admin_permissions() or (
        (
            permissions.has_manager_permissions()
            or entity["created_by"] == persons_service.get_current_user()["id"]
        )
        and check_belong_to_project(entity["project_id"])
    ):
        is_allowed = True
    elif permissions.has_supervisor_permissions() and check_belong_to_project(
        entity["project_id"]
    ):
        # checks that the supervisor only modifies columns
        # for which he is authorized
        allowed_columns = set(["data"])
        if len(set(new_data.keys()) - allowed_columns) == 0:
            user_departments = persons_service.get_current_user(
                relations=True
            )["departments"]
            if user_departments == []:
                is_allowed = True
            else:
                entity_type = None
                if shots_service.is_shot(entity):
                    entity_type = "Shot"
                elif assets_service.is_asset(
                    entities_service.get_entity_raw(entity["id"])
                ):
                    entity_type = "Asset"
                elif edits_service.is_edit(entity):
                    entity_type = "Edit"
                if entity_type:
                    descriptors = [
                        descriptor
                        for descriptor in projects_service.get_metadata_descriptors(
                            entity["project_id"]
                        )
                        if descriptor["entity_type"] == entity_type
                    ]
                    found_and_in_departments = False
                    for descriptor_name in new_data["data"].keys():
                        found_and_in_departments = False
                        for descriptor in descriptors:
                            if descriptor["field_name"] == descriptor_name:
                                found_and_in_departments = (
                                    len(
                                        set(descriptor["departments"])
                                        & set(user_departments)
                                    )
                                    > 0
                                )
                                break
                        if not found_and_in_departments:
                            break
                    if found_and_in_departments:
                        is_allowed = True

    if not is_allowed:
        raise permissions.PermissionDenied
    return is_allowed


def check_task_department_access(task_id, person_id):
    """
    Return true if current user is an admin or is a manager and is in team
    or is a supervisor in the department of the task or is an artist assigning
    himself in the department of the task.
    """
    user = persons_service.get_current_user(relations=True)
    task = tasks_service.get_task(task_id)
    task_type = tasks_service.get_task_type(task["task_type_id"])
    is_allowed = permissions.has_admin_permissions() or (
        check_belong_to_project(task["project_id"])
        and (
            permissions.has_manager_permissions()
            or (
                permissions.has_supervisor_permissions()
                and (
                    user["departments"] == []
                    or (
                        task_type["department_id"] in user["departments"]
                        and len(
                            set(
                                persons_service.get_person(person_id)[
                                    "departments"
                                ]
                            )
                            & set(user["departments"])
                        )
                        > 0
                    )
                )
            )
            or (
                task_type["department_id"] in user["departments"]
                and person_id == user["id"]
            )
        )
    )
    if not is_allowed:
        raise permissions.PermissionDenied
    return is_allowed


def check_person_is_not_bot(person_id):
    """
    Return true if person is not a bot else raise PermissionDenied
    """
    if persons_service.get_person(person_id)["is_bot"]:
        raise permissions.PermissionDenied
    else:
        return True


def check_task_department_access_for_unassign(task_id, person_id=None):
    """
    Return true if current user is an admin or is a manager and is in team
    or is a supervisor in the department of the task or is an artist assigning
    himself in the department of the task.
    """
    user = persons_service.get_current_user(relations=True)
    task = tasks_service.get_task(task_id)
    task_type = tasks_service.get_task_type(task["task_type_id"])
    is_allowed = permissions.has_admin_permissions() or (
        check_belong_to_project(task["project_id"])
        and (
            permissions.has_manager_permissions()
            or (
                permissions.has_supervisor_permissions()
                and (
                    user["departments"] == []
                    or task_type["department_id"] in user["departments"]
                )
            )
            or (user["id"] in task["assignees"] and person_id == user["id"])
        )
    )
    if not is_allowed:
        raise permissions.PermissionDenied
    return is_allowed


def check_all_departments_access(project_id, departments=[]):
    """
    Return true if current user is admin or is manager and is in team or is
    supervisor and is in team and have access to all departments.
    """
    if not isinstance(departments, list):
        departments = [departments]
    is_allowed = False
    if permissions.has_admin_permissions() or (
        permissions.has_manager_permissions()
        and check_belong_to_project(project_id)
    ):
        is_allowed = True
    elif permissions.has_supervisor_permissions() and check_belong_to_project(
        project_id
    ):
        user_departments = persons_service.get_current_user(relations=True)[
            "departments"
        ]
        is_allowed = departments and (
            user_departments == []
            or all(
                department in departments for department in user_departments
            )
        )
    if not is_allowed:
        raise permissions.PermissionDenied
    return is_allowed


def check_playlist_access(playlist, supervisor_access=False):
    check_project_access(playlist["project_id"])
    is_manager = permissions.has_manager_permissions()
    is_client = permissions.has_client_permissions()
    has_supervisor_access = (
        supervisor_access and permissions.has_supervisor_permissions()
    )
    has_client_access = is_client and playlist["for_client"]
    is_allowed = is_manager or has_client_access or has_supervisor_access
    if not is_allowed:
        raise permissions.PermissionDenied
    return True


def check_day_off_access(day_off):
    """
    Return true if current user is admin or day_off is for itself
    """
    user = persons_service.get_current_user()
    is_admin = permissions.has_admin_permissions()
    is_same_person = user["id"] == day_off["person_id"]
    if not (is_admin or is_same_person):
        raise permissions.PermissionDenied
    return True


def get_filters():
    """
    Retrieve search filters used by current user. It groups them by
    list type and project_id. If the filter is not related to a project,
    the project_id is all.
    """
    current_user = persons_service.get_current_user()
    return get_user_filters(current_user["id"])


@cache.memoize_function(120)
def get_user_filters(current_user_id):
    """
    Retrieve search filters used for given user. It groups them by
    list type and project_id. If the filter is not related to a project,
    the project_id is all.
    """

    result = {}

    filters = (
        SearchFilter.query.outerjoin(Project)
        .outerjoin(ProjectStatus)
        .filter(
            or_(
                SearchFilter.person_id == current_user_id,
                SearchFilter.is_shared == True,
            )
        )
        .filter(
            or_(build_open_project_filter(), SearchFilter.project_id == None)
        )
        .all()
    )

    current_user = persons_service.get_current_user(relations=True)
    is_manager = permissions.has_manager_permissions()

    for search_filter in filters:
        department_id = search_filter.department_id
        is_in_departments = (
            department_id is not None
            and str(department_id) in current_user["departments"]
        )

        if department_id is None or is_manager or is_in_departments:
            if search_filter.list_type not in result:
                result[search_filter.list_type] = {}
            subresult = result[search_filter.list_type]

            if search_filter.project_id is None:
                project_id = "all"
            else:
                project_id = str(search_filter.project_id)

            if project_id not in subresult:
                subresult[project_id] = []

            subresult[project_id].append(search_filter.serialize())

    return result


def create_filter(
    list_type,
    name,
    query,
    project_id=None,
    entity_type=None,
    is_shared=False,
    search_filter_group_id=None,
    department_id=None,
):
    """
    Add a new search filter to the database.
    """
    current_user = persons_service.get_current_user()

    if project_id is None or (
        project_id is not None and not has_manager_project_access(project_id)
    ):
        is_shared = False

    if search_filter_group_id is not None:
        search_filter_group = SearchFilterGroup.get_by(
            id=search_filter_group_id
        )
        if search_filter_group is None:
            raise SearchFilterGroupNotFoundException
        if is_shared != search_filter_group.is_shared:
            raise WrongParameterException(
                "A search filter should have the same value for is_shared than its search filter group."
            )

    if department_id is not None:
        department = tasks_service.get_department(department_id)
        if department is None:
            raise WrongParameterException(
                f"No department found with id: {department_id}"
            )

    search_filter = SearchFilter.create(
        list_type=list_type,
        name=name,
        search_query=query,
        project_id=project_id,
        person_id=current_user["id"],
        entity_type=entity_type,
        is_shared=is_shared,
        search_filter_group_id=search_filter_group_id,
        department_id=department_id,
    )
    search_filter.serialize()
    if search_filter.is_shared:
        clear_filter_cache()
    else:
        clear_filter_cache(current_user["id"])
    return search_filter.serialize()


def update_filter(search_filter_id, data):
    """
    Update given filter from database.
    """
    current_user = persons_service.get_current_user()
    search_filter = SearchFilter.get_by(
        id=search_filter_id, person_id=current_user["id"]
    )
    if current_user["role"] == "admin" and search_filter is None:
        search_filter = SearchFilter.get_by(id=search_filter_id)

    if search_filter is None:
        raise SearchFilterNotFoundException

    department_id = data.get("department_id", None)
    if department_id is not None:
        department = tasks_service.get_department(department_id)
        if department is None:
            raise WrongParameterException(
                f"No department found with id: {department_id}"
            )

    if (
        data.get("is_shared", None) is not None
        and search_filter.is_shared != data["is_shared"]
        and (
            data.get("project_id", None) is None
            or (
                data["project_id"] is not None
                and not has_manager_project_access(data["project_id"])
            )
        )
    ):
        data["is_shared"] = False

    if (
        search_filter_group_id := data.get(
            "search_filter_group_id", search_filter.search_filter_group_id
        )
    ) is not None:
        search_filter_group = SearchFilterGroup.get_by(
            id=search_filter_group_id
        )
        if search_filter_group is None:
            raise SearchFilterGroupNotFoundException
        if (
            data.get("is_shared", search_filter.is_shared)
            != search_filter_group.is_shared
        ):
            raise WrongParameterException(
                "A search filter should have the same value for is_shared than its search filter group."
            )

    department_id = data.get("department_id", None)
    if department_id is not None:
        department = tasks_service.get_department(department_id)
        if department is None:
            raise WrongParameterException(
                f"No department found with id: {department_id}"
            )

    search_filter.update(data)
    if search_filter.is_shared:
        clear_filter_cache()
    else:
        clear_filter_cache(current_user["id"])
    return search_filter.serialize()


def remove_filter(search_filter_id):
    """
    Remove given filter from database.
    """
    current_user = persons_service.get_current_user()
    search_filter = SearchFilter.get_by(
        id=search_filter_id, person_id=current_user["id"]
    )
    if search_filter is None:
        raise SearchFilterNotFoundException
    search_filter.delete()
    if search_filter.is_shared:
        clear_filter_cache()
    else:
        clear_filter_cache(current_user["id"])
    return search_filter.serialize()


def get_filter_groups():
    """
    Retrieve search filter groups used by current user. It groups them by
    list type and project_id. If the filter group is not related to a project,
    the project_id is all.
    """
    current_user = persons_service.get_current_user()
    return get_user_filter_groups(current_user["id"])


@cache.memoize_function(10)
def get_user_filter_groups(current_user_id):
    """
    Retrieve search filter groups used for given user. It groups them by
    list type and project_id. If the filter group is not related to a project,
    the project_id is all.
    """

    result = {}

    filter_groups = (
        SearchFilterGroup.query.outerjoin(
            Project, Project.id == SearchFilterGroup.project_id
        )
        .outerjoin(
            ProjectStatus, ProjectStatus.id == Project.project_status_id
        )
        .filter(
            or_(
                SearchFilterGroup.person_id == current_user_id,
                SearchFilterGroup.is_shared == True,
            )
        )
        .filter(or_(build_open_project_filter(), Project.id == None))
        .order_by(SearchFilterGroup.created_at.desc())
        .all()
    )

    current_user = persons_service.get_current_user(relations=True)
    is_manager = permissions.has_manager_permissions()

    for search_filter_group in filter_groups:
        if search_filter_group.list_type not in result:
            result[search_filter_group.list_type] = {}

        department_id = search_filter_group.department_id
        is_in_departments = (
            department_id is not None
            and str(department_id) in current_user["departments"]
        )
        if department_id is None or is_manager or is_in_departments:
            subresult = result[search_filter_group.list_type]

            if search_filter_group.project_id is None:
                project_id = "all"
            else:
                project_id = str(search_filter_group.project_id)

            if project_id not in subresult:
                subresult[project_id] = []
            subresult[project_id].append(search_filter_group.serialize())

    return result


def create_filter_group(
    list_type,
    name,
    color,
    project_id=None,
    entity_type=None,
    is_shared=False,
    department_id=None,
):
    """
    Add a new search filter group to the database.
    """
    current_user = persons_service.get_current_user()
    if project_id is None or (
        project_id is not None and not has_manager_project_access(project_id)
    ):
        is_shared = False

    if department_id is not None:
        department = tasks_service.get_department(department_id)
        if department is None:
            raise WrongParameterException(
                f"No department found with id: {department_id}"
            )

    search_filter_group = SearchFilterGroup.create(
        list_type=list_type,
        name=name,
        color=color,
        project_id=project_id,
        person_id=current_user["id"],
        entity_type=entity_type,
        is_shared=is_shared,
        department_id=department_id,
    )
    search_filter_group.serialize()

    if search_filter_group.is_shared:
        clear_filter_group_cache()
    else:
        clear_filter_group_cache(current_user["id"])

    return search_filter_group.serialize()


def get_filter_group(search_filter_group_id):
    """
    Get given filter group from the database.
    """
    current_user = persons_service.get_current_user()
    search_filter_group = SearchFilterGroup.get_by(
        id=search_filter_group_id, person_id=current_user["id"]
    )
    if search_filter_group is None:
        raise SearchFilterGroupNotFoundException
    return search_filter_group.serialize()


def update_filter_group(search_filter_group_id, data):
    """
    Update given filter group from database.
    """
    current_user = persons_service.get_current_user()
    search_filter_group = SearchFilterGroup.get_by(
        id=search_filter_group_id, person_id=current_user["id"]
    )

    if search_filter_group is None:
        raise SearchFilterGroupNotFoundException

    if (
        data.get("is_shared", None) is not None
        and search_filter_group.is_shared != data["is_shared"]
        and (
            data.get("project_id", None) is None
            or (
                data["project_id"] is not None
                and not has_manager_project_access(data["project_id"])
            )
        )
    ):
        data["is_shared"] = False

    search_filter_group.update(data)

    if (
        data.get("is_shared", None) is not None
        and data.get("project_id", None) is not None
        and has_manager_project_access(data["project_id"])
    ):
        if (
            SearchFilter.query.filter_by(
                search_filter_group_id=search_filter_group_id
            ).update({"is_shared": data["is_shared"]})
            > 0
        ):
            SearchFilter.query.session.commit()
            clear_filter_cache()

    if search_filter_group.is_shared:
        clear_filter_group_cache()
    else:
        clear_filter_group_cache(current_user["id"])
    return search_filter_group.serialize()


def remove_filter_group(search_filter_group_id):
    """
    Remove given filter group from database.
    """
    current_user = persons_service.get_current_user()
    search_filter_group = SearchFilterGroup.get_by(
        id=search_filter_group_id, person_id=current_user["id"]
    )
    if search_filter_group is None:
        raise SearchFilterGroupNotFoundException
    if (
        SearchFilter.query.filter_by(
            search_filter_group_id=search_filter_group_id
        ).delete()
        > 0
    ):
        SearchFilter.query.session.commit()
        clear_filter_cache()
    search_filter_group.delete()
    if search_filter_group.is_shared:
        clear_filter_group_cache()
    else:
        clear_filter_group_cache(current_user["id"])
    return search_filter_group.serialize()


def get_notification(notification_id):
    """
    Return notification matching given ID as a dictionnary.
    """
    notifications = get_last_notifications(notification_id)

    if len(notifications) == 0:
        raise NotificationNotFoundException

    return notifications[0]


def update_notification(notification_id, read):
    """
    Update read status of given notification.
    """
    current_user = persons_service.get_current_user()
    notification = Notification.get_by(
        id=notification_id, person_id=current_user["id"]
    )
    notification.update({"read": read})
    if read:
        events.emit(
            "notification:read",
            {
                "person_id": current_user["id"],
                "notification_id": notification_id,
            },
        )
    else:
        events.emit(
            "notification:unread",
            {
                "person_id": current_user["id"],
                "notification_id": notification_id,
            },
        )
    return notification.serialize()


def get_unread_notifications_count(notification_id=None):
    """
    Return the number of unread notifications.
    """
    current_user = persons_service.get_current_user()
    return Notification.query.filter_by(
        person_id=current_user["id"], read=False
    ).count()


def get_last_notifications(
    notification_id=None,
    after=None,
    before=None,
    task_type_id=None,
    task_status_id=None,
    notification_type=None,
    read=None,
    watching=None,
):
    """
    Return last 100 user notifications.
    """
    current_user = persons_service.get_current_user()
    Author = aliased(Person, name="author")
    is_current_user_artist = current_user["role"] == "user"
    result = []
    query = (
        Notification.query.filter_by(person_id=current_user["id"])
        .order_by(Notification.created_at.desc())
        .join(Author, Author.id == Notification.author_id)
        .join(Task, Task.id == Notification.task_id)
        .join(Project, Project.id == Task.project_id)
        .outerjoin(
            Subscription,
            and_(
                Subscription.task_id == Task.id,
                Subscription.person_id == current_user["id"],
            ),
        )
        .outerjoin(Comment, Comment.id == Notification.comment_id)
        .add_columns(
            Project.id,
            Project.name,
            Task.task_type_id,
            Comment.id,
            Comment.task_status_id,
            Comment.text,
            Comment.replies,
            Task.entity_id,
            Subscription.id,
            Author.role,
        )
    )

    if notification_id is not None:
        query = query.filter(Notification.id == notification_id)

    if after is not None:
        query = query.filter(
            Notification.created_at
            > func.cast(after, Notification.created_at.type)
        )

    if before is not None:
        query = query.filter(
            Notification.created_at
            < func.cast(before, Notification.created_at.type)
        )

    if task_type_id is not None:
        query = query.filter(Task.task_type_id == task_type_id)

    if task_status_id is not None:
        query = query.filter(Task.task_status_id == task_status_id)

    if notification_type is not None:
        query = query.filter(Notification.type == notification_type)

    if read is not None:
        query = query.filter(Notification.read == read)

    if watching is not None:
        if watching:
            query = query.filter(Subscription.id != None)
        else:
            query = query.filter(Subscription.id == None)

    notifications = query.limit(100).all()

    for (
        notification,
        project_id,
        project_name,
        task_type_id,
        comment_id,
        task_status_id,
        comment_text,
        comment_replies,
        task_entity_id,
        subscription_id,
        role,
    ) in notifications:
        (full_entity_name, episode_id, entity_preview_file_id) = (
            names_service.get_full_entity_name(task_entity_id)
        )
        preview_file_id = None
        mentions = []
        department_mentions = []
        reply_mentions = []
        reply_department_mentions = []
        if comment_id is not None:
            comment = Comment.get(comment_id)
            if len(comment.previews) > 0:
                preview_file_id = comment.previews[0].id
            mentions = comment.mentions or []
            department_mentions = comment.department_mentions or []

        reply_text = ""
        if notification.type in ["reply", "reply-mention"]:
            reply = next(
                (
                    reply
                    for reply in comment_replies
                    if reply["id"] == str(notification.reply_id)
                ),
                None,
            )
            if reply is not None:
                reply_text = reply["text"]
                reply_mentions = reply.get("mentions", []) or []
                reply_department_mentions = (
                    reply.get("department_mentions", []) or []
                )
            else:
                reply_mentions = []
                reply_department_mentions = []

        if role == "client" and is_current_user_artist:
            comment_text = ""
            reply_text = ""

        result.append(
            fields.serialize_dict(
                {
                    "id": notification.id,
                    "type": "Notification",
                    "notification_type": notification.type,
                    "author_id": notification.author_id,
                    "comment_id": notification.comment_id,
                    "task_id": notification.task_id,
                    "task_type_id": task_type_id,
                    "task_status_id": task_status_id,
                    "mentions": mentions,
                    "department_mentions": department_mentions,
                    "reply_mentions": reply_mentions,
                    "reply_department_mentions": reply_department_mentions,
                    "preview_file_id": preview_file_id,
                    "project_id": project_id,
                    "project_name": project_name,
                    "comment_text": comment_text,
                    "reply_text": reply_text,
                    "created_at": notification.created_at,
                    "read": notification.read,
                    "change": notification.change,
                    "full_entity_name": full_entity_name,
                    "episode_id": episode_id,
                    "entity_preview_file_id": entity_preview_file_id,
                    "subscription_id": subscription_id,
                }
            )
        )

    return result


def mark_notifications_as_read():
    """
    Mark all recent notifications for current_user as read. It is useful
    to mark a list of notifications as read after an user retrieved them.
    """
    from sqlalchemy import update
    from zou.app import db

    current_user = persons_service.get_current_user()
    update_stmt = (
        update(Notification)
        .where(Notification.person_id == current_user["id"])
        .where(Notification.read == False)
        .values(read=True)
    )

    db.session.execute(update_stmt)
    db.session.commit()
    events.emit("notification:all-read", {"person_id": current_user["id"]})
    return True


def has_task_subscription(task_id):
    """
    Returns true if a subscription entry exists for current user and given
    task.
    """
    current_user = persons_service.get_current_user()
    return notifications_service.has_task_subscription(
        current_user["id"], task_id
    )


def subscribe_to_task(task_id):
    """
    Create a subscription entry for current user and given task
    """
    current_user = persons_service.get_current_user()
    return notifications_service.subscribe_to_task(current_user["id"], task_id)


def unsubscribe_from_task(task_id):
    """
    Remove subscription entry for current user and given task
    """
    current_user = persons_service.get_current_user()
    return notifications_service.unsubscribe_from_task(
        current_user["id"], task_id
    )


def has_sequence_subscription(sequence_id, task_type_id):
    """
    Returns true if a subscription entry exists for current user and given
    sequence.
    """
    current_user = persons_service.get_current_user()
    return notifications_service.has_sequence_subscription(
        current_user["id"], sequence_id, task_type_id
    )


def subscribe_to_sequence(sequence_id, task_type_id):
    """
    Create a subscription entry for current user and given sequence
    """
    current_user = persons_service.get_current_user()
    return notifications_service.subscribe_to_sequence(
        current_user["id"], sequence_id, task_type_id
    )


def unsubscribe_from_sequence(sequence_id, task_type_id):
    """
    Remove subscription entry for current user and given sequence
    """
    current_user = persons_service.get_current_user()
    return notifications_service.unsubscribe_from_sequence(
        current_user["id"], sequence_id, task_type_id
    )


def get_sequence_subscriptions(project_id, task_type_id):
    """
    Return list of sequence ids for which the current user has subscriptions
    for given project and task type.
    """
    current_user = persons_service.get_current_user()
    return notifications_service.get_all_sequence_subscriptions(
        current_user["id"], project_id, task_type_id
    )


def get_timezone():
    try:
        timezone = persons_service.get_current_user()["timezone"]
    except Exception:
        timezone = config.DEFAULT_TIMEZONE
    return timezone or config.DEFAULT_TIMEZONE


def get_context():
    context = {
        "asset_types": assets_service.get_asset_types(),
        "custom_actions": custom_actions_service.get_custom_actions(),
        "status_automations": status_automations_service.get_status_automations(),
        "departments": tasks_service.get_departments(),
        "studios": tasks_service.get_studios(),
        "notification_count": get_unread_notifications_count(),
        "persons": persons_service.get_persons(
            minimal=not permissions.has_manager_permissions()
        ),
        "project_status": projects_service.get_project_statuses(),
        "projects": get_open_projects(),
        "task_types": tasks_service.get_task_types(),
        "task_status": tasks_service.get_task_statuses(),
        "search_filters": get_filters(),
        "search_filter_groups": get_filter_groups(),
        "preview_background_files": files_service.get_preview_background_files(),
    }

    if permissions.has_admin_permissions():
        context["user_limit"] = config.USER_LIMIT
    return context



import urllib.parse

from flask import request, jsonify, current_app, redirect, make_response
from flask_restful import Resource
from flask_principal import (
    Identity,
    AnonymousIdentity,
    identity_changed,
)
from flask_jwt_extended import (
    jwt_required,
    create_access_token,
    create_refresh_token,
    set_access_cookies,
    set_refresh_cookies,
    unset_jwt_cookies,
    unset_refresh_cookies,
    get_jwt,
)

from sqlalchemy.exc import OperationalError, TimeoutError
from babel.dates import format_datetime
from saml2 import entity, client_base

from zou.app import app, config
from zou.app.mixin import ArgsMixin
from zou.app.utils import auth, emails, permissions, date_helpers
from zou.app.services import (
    persons_service,
    auth_service,
    events_service,
)

from zou.app.utils.flask import is_from_browser
from zou.app.utils.saml import saml_client_for

from zou.app.stores import auth_tokens_store
from zou.app.services.exception import (
    EmailOTPAlreadyEnabledException,
    EmailOTPNotEnabledException,
    FIDONoPreregistrationException,
    FIDONotEnabledException,
    FIDOServerException,
    MissingOTPException,
    NoAuthStrategyConfigured,
    NoTwoFactorAuthenticationEnabled,
    PersonNotFoundException,
    TooMuchLoginFailedAttemps,
    TOTPAlreadyEnabledException,
    TOTPNotEnabledException,
    UnactiveUserException,
    UserCantConnectDueToNoFallback,
    WrongOTPException,
    WrongPasswordException,
    WrongUserException,
)


class AuthenticatedResource(Resource):
    """
    Returns information if the user is authenticated else it returns a 401
    response.
    It can be used by third party tools, especially browser frontend, to know
    if current user is still logged in.
    """

    @jwt_required()
    def get(self):
        """
        Returns information if the user is authenticated else it returns a 401
        response.
        ---
        description:  It can be used by third party tools, especially browser frontend, to know if current user is still logged in.
        tags:
            - Authentication
        responses:
          200:
            description: User authenticated
          401:
            description: Person not found
        """
        person = persons_service.get_current_user(relations=True)
        organisation = persons_service.get_organisation(
            sensitive=permissions.has_admin_permissions()
        )
        return {
            "authenticated": True,
            "user": person,
            "organisation": organisation,
        }


class LogoutResource(Resource):
    """
    Log user out by revoking his auth tokens. Once log out, current user
    cannot access to API anymore.
    """

    @jwt_required()
    @permissions.require_person
    def get(self):
        """
        Log user out by revoking his auth tokens.
        ---
        description: Once logged out, current user cannot access the API anymore.
        tags:
            - Authentication
        responses:
          200:
            description: Logout successful
          500:
            description: Access token not found
        """
        try:
            auth_service.logout(get_jwt()["jti"])
            identity_changed.send(
                current_app._get_current_object(), identity=AnonymousIdentity()
            )
        except KeyError:
            return {"Access token not found."}, 500

        logout_data = {"logout": True}

        if is_from_browser(request.user_agent):
            response = jsonify(logout_data)
            unset_jwt_cookies(response)
            return response
        else:
            return logout_data


class LoginResource(Resource, ArgsMixin):
    """
    Log in user by creating and registering auth tokens. Login is based
    on email and password. If no user match given email and a destkop ID,
    it looks in matching the desktop ID with the one stored in database. It is
    useful for clients that run on desktop tools and that don't know user
    email.
    """

    def post(self):
        """
        Log in user by creating and registering auth tokens.
        ---
        description: Login is based on email and password.
                     If no user match given email and a destkop ID, it looks in matching the desktop ID with the one stored in database.
                     It is useful for clients that run on desktop tools and that don't know user email.
        tags:
            - Authentication
        parameters:
          - in: formData
            name: email
            required: True
            type: string
            format: email
            example: admin@example.com
          - in: formData
            name: password
            required: True
            type: string
            format: password
            example: mysecretpassword
          - in: formData
            name: otp
            required: False
            type: string
            format: password
            example: 123456
        responses:
          200:
            description: Login successful
          400:
            description: Login failed
          500:
            description: Database not reachable
        """
        (
            email,
            password,
            totp,
            email_otp,
            fido_authentication_response,
            recovery_code,
        ) = self.get_arguments()
        try:
            user = auth_service.check_auth(
                app,
                email,
                password,
                totp,
                email_otp,
                fido_authentication_response,
                recovery_code,
            )

            if auth_service.is_default_password(app, password):
                token = auth_service.generate_reset_token()
                auth_tokens_store.add(
                    "reset-token-%s" % email, token, ttl=3600 * 2
                )
                current_app.logger.info(
                    "User %s must change his password." % email
                )
                return (
                    {
                        "login": False,
                        "default_password": True,
                        "token": token,
                    },
                    400,
                )

            access_token = create_access_token(
                identity=user["id"],
                additional_claims={
                    "identity_type": "person",
                },
            )
            refresh_token = create_refresh_token(
                identity=user["id"],
                additional_claims={
                    "identity_type": "person",
                },
            )
            identity_changed.send(
                current_app._get_current_object(),
                identity=Identity(user["id"], "person"),
            )

            ip_address = request.environ.get(
                "HTTP_X_REAL_IP", request.remote_addr
            )

            organisation = persons_service.get_organisation(
                sensitive=user["role"] != "admin"
            )

            response = jsonify(
                {
                    "user": user,
                    "organisation": organisation,
                    "login": True,
                    "access_token": access_token,
                    "refresh_token": refresh_token,
                }
            )

            if is_from_browser(request.user_agent):
                set_access_cookies(response, access_token)
                set_refresh_cookies(response, refresh_token)
                events_service.create_login_log(user["id"], ip_address, "web")
            else:
                events_service.create_login_log(
                    user["id"], ip_address, "script"
                )
            current_app.logger.info(f"User {email} is logged in.")
            return response
        except WrongUserException:
            current_app.logger.info(f"User {email} is not registered.")
            return {"login": False}, 400
        except WrongPasswordException:
            current_app.logger.info(f"User {email} gave a wrong password.")
            return {"login": False}, 400
        except NoAuthStrategyConfigured:
            current_app.logger.info(
                "Authentication strategy is not properly configured."
            )
            return {"login": False}, 409
        except UserCantConnectDueToNoFallback:
            current_app.logger.info(
                f"User {email} can't login due to no fallback from LDAP."
            )
            return {"login": False}, 400
        except TimeoutError:
            current_app.logger.info("Timeout occurs while logging in.")
            return {"login": False}, 400
        except UnactiveUserException:
            current_app.logger.info(f"User {email} is unactive.")
            return (
                {
                    "error": True,
                    "login": False,
                    "message": "User is unactive, he cannot log in.",
                },
                401,
            )
        except TooMuchLoginFailedAttemps:
            current_app.logger.info(
                f"User {email} can't log in due to too much login failed attemps."
            )
            return (
                {
                    "error": True,
                    "login": False,
                    "too_many_failed_login_attemps": True,
                },
                400,
            )
        except MissingOTPException as e:
            current_app.logger.info(
                f"User {email} can't log in due to missing OTP."
            )
            return (
                {
                    "error": True,
                    "login": False,
                    "missing_OTP": True,
                    "preferred_two_factor_authentication": e.preferred_two_factor_authentication,
                    "two_factor_authentication_enabled": e.two_factor_authentication_enabled,
                },
                400,
            )
        except WrongOTPException:
            current_app.logger.info(
                f"User {email} can't log in due to wrong OTP."
            )
            return (
                {
                    "error": True,
                    "login": False,
                    "wrong_OTP": True,
                },
                400,
            )
        except OperationalError as exception:
            current_app.logger.error(exception, exc_info=1)
            return (
                {
                    "error": True,
                    "login": False,
                    "message": "Database doesn't seem reachable.",
                },
                500,
            )
        except Exception as exception:
            current_app.logger.error(exception, exc_info=1)
            if hasattr(exception, "message"):
                message = exception.message
            else:
                message = str(exception)
            return {"error": True, "login": False, "message": message}, 500

    def get_arguments(self):
        args = self.get_args(
            [
                {
                    "name": "email",
                    "required": True,
                    "help": "User email is missing.",
                },
                ("password", "default"),
                "totp",
                "email_otp",
                ("fido_authentication_response", None, False, dict),
                "recovery_code",
            ]
        )

        return (
            args["email"],
            args["password"],
            args["totp"],
            args["email_otp"],
            args["fido_authentication_response"],
            args["recovery_code"],
        )


class RefreshTokenResource(Resource):
    @jwt_required(refresh=True)
    @permissions.require_person
    def get(self):
        """
        Tokens are considered as outdated every two weeks.
        ---
        description: This route allows to make their lifetime long before they get outdated.
        tags:
            - Authentication
        responses:
          200:
            description: Access Token
        """
        user = persons_service.get_current_user()
        access_token = create_access_token(
            identity=user["id"],
            additional_claims={
                "identity_type": "person",
            },
        )
        if is_from_browser(request.user_agent):
            response = jsonify({"refresh": True})
            set_access_cookies(response, access_token)
            unset_refresh_cookies(response)
        else:
            return {"access_token": access_token}


class RegistrationResource(Resource, ArgsMixin):
    """
    Allow a user to register himself to the service.
    """

    def post(self):
        """
        Allow a user to register himself to the service.
        ---
        tags:
            - Authentication
        parameters:
          - in: formData
            name: email
            required: True
            type: string
            format: email
            example: admin@example.com
          - in: formData
            name: password
            required: True
            type: string
            format: password
          - in: formData
            name: password_2
            required: True
            type: string
            format: password
          - in: formData
            name: first_name
            required: True
            type: string
          - in: formData
            name: last_name
            required: True
            type: string
        responses:
          201:
            description: Registration successful
          400:
            description: Invalid password or email
        """
        (
            email,
            password,
            password_2,
            first_name,
            last_name,
        ) = self.get_arguments()

        try:
            email = auth.validate_email(email)
            auth.validate_password(password, password_2)
            password = auth.encrypt_password(password)
            persons_service.create_person(
                email, password, first_name, last_name
            )
            return {"registration_success": True}, 201
        except auth.PasswordsNoMatchException:
            return (
                {
                    "error": True,
                    "message": "Confirmation password doesn't match.",
                },
                400,
            )
        except auth.PasswordTooShortException:
            return {"error": True, "message": "Password is too short."}, 400
        except auth.EmailNotValidException as exception:
            return {"error": True, "message": str(exception)}, 400

    def get_arguments(self):
        args = self.get_args(
            [
                {
                    "name": "email",
                    "required": True,
                    "help": "User email is missing.",
                },
                {
                    "name": "first_name",
                    "required": True,
                    "help": "First name is missing.",
                },
                {
                    "name": "last_name",
                    "required": True,
                    "help": "Last name is missing.",
                },
                {
                    "name": "password",
                    "required": True,
                    "help": "Password is missing.",
                },
                {
                    "name": "password_2",
                    "required": True,
                    "help": "Confirmation password is missing.",
                },
            ]
        )

        return (
            args["email"],
            args["password"],
            args["password_2"],
            args["first_name"],
            args["last_name"],
        )


class ChangePasswordResource(Resource, ArgsMixin):
    """
    Allow the user to change his password. Prior to modify the password,
    it requires to give the current password (to make sure the user changing
    the password is not someone who stealed the session).
    The new password requires a confirmation to ensure that the user didn't
    make mistake by typing his new password.
    """

    @jwt_required()
    @permissions.require_person
    def post(self):
        """
        Allow the user to change his password.
        ---
        description: Prior to modifying the password, it requires to give the current password
                     (to make sure the user changing the password is not someone who stealed the session).
                     The new password requires a confirmation to ensure that the user didn't
                     make a mistake by typing his new password.
        tags:
            - Authentication
        parameters:
          - in: formData
            name: old_password
            required: True
            type: string
            format: password
          - in: formData
            name: password
            required: True
            type: string
            format: password
          - in: formData
            name: password_2
            required: True
            type: string
            format: password
        responses:
          200:
            description: Password changed
          400:
            description: Invalid password or inactive user
        """
        (old_password, password, password_2) = self.get_arguments()

        try:
            user = persons_service.get_current_user()
            auth_service.check_auth(
                app, user["email"], old_password, no_otp=True
            )
            auth.validate_password(password, password_2)
            password = auth.encrypt_password(password)
            persons_service.update_password(user["email"], password)
            current_app.logger.info(
                "User %s has changed his password" % user["email"]
            )
            organisation = persons_service.get_organisation()
            time_string = format_datetime(
                date_helpers.get_utc_now_datetime(),
                tzinfo=user["timezone"],
                locale=user["locale"],
            )
            person_IP = request.headers.get("X-Forwarded-For", None)
            html = f"""<p>Hello {user["first_name"]},</p>

<p>
You have successfully changed your password at this date : {time_string}.

Your IP when you have changed your password is : {person_IP}.
</p>

Thank you and see you soon on Kitsu,
</p>
<p>
{organisation["name"]} Team
</p>
"""
            subject = f"{organisation['name']} - Kitsu: password changed"
            emails.send_email(subject, html, user["email"])
            return {"success": True}

        except auth.PasswordsNoMatchException:
            return (
                {
                    "error": True,
                    "message": "Confirmation password doesn't match.",
                },
                400,
            )
        except auth.PasswordTooShortException:
            return {"error": True, "message": "Password is too short."}, 400
        except UnactiveUserException:
            return {"error": True, "message": "User is unactive."}, 400
        except WrongPasswordException:
            return {"error": True, "message": "Old password is wrong."}, 400

    def get_arguments(self):
        args = self.get_args(
            [
                {
                    "name": "old_password",
                    "required": True,
                    "help": "Old password is missing.",
                },
                {
                    "name": "password",
                    "required": True,
                    "help": "New password is missing.",
                },
                {
                    "name": "password_2",
                    "required": True,
                    "help": "New password confirmation is missing.",
                },
            ]
        )

        return (args["old_password"], args["password"], args["password_2"])


class ResetPasswordResource(Resource, ArgsMixin):
    """
    Resource to allow a user to change his password when he forgets it.
    It uses a classic scheme: a token is sent by email to the user. Then
    he can change his password.
    """

    def put(self):
        """
        Resource to allow a user to change his password when he forgets it.
        ---
        description: "It uses a classic scheme: a token is sent by email to the user.
                     Then he can change his password."
        tags:
            - Authentication
        parameters:
          - in: formData
            name: email
            required: True
            type: string
            format: email
            example: admin@example.com
          - in: formData
            name: token
            required: True
            type: string
            format: JWT token
          - in: formData
            name: password
            required: True
            type: string
            format: password
          - in: formData
            name: password2
            required: True
            type: string
            format: password
        responses:
          200:
            description: Password reset
          400:
            description: Invalid password
                         Wrong or expired token
                         Inactive user
        """
        args = self.get_args(
            [
                ("email", "", True),
                ("token", "", True),
                ("password", "", True),
                ("password2", "", True),
            ]
        )

        try:
            token_from_store = auth_tokens_store.get(
                "reset-token-%s" % args["email"]
            )
            if token_from_store == args["token"]:
                auth.validate_password(args["password"], args["password2"])
                password = auth.encrypt_password(args["password"])
                persons_service.update_password(args["email"], password)
                auth_tokens_store.delete("reset-token-%s" % args["email"])
                current_app.logger.info(
                    "User %s has reset his password" % args["email"]
                )
                return {"success": True}
            else:
                return (
                    {"error": True, "message": "Wrong or expired token."},
                    400,
                )

        except auth.PasswordsNoMatchException:
            return (
                {
                    "error": True,
                    "message": "Confirmation password doesn't match.",
                },
                400,
            )
        except auth.PasswordTooShortException:
            return {"error": True, "message": "Password is too short."}, 400
        except UnactiveUserException:
            return {"error": True, "message": "User is inactive."}, 400

    def post(self):
        """
        Resource to allow a user to change his password when he forgets it.
        ---
        description: "It uses a classic scheme: a token is sent by email to the user.
                     Then he can change his password."
        tags:
            - Authentication
        parameters:
          - in: formData
            name: email
            required: True
            type: string
            format: email
            example: admin@example.com
        responses:
          200:
            description: Reset token sent
          400:
            description: Email not listed in database
        """
        args = self.get_args([("email", "", True)])

        try:
            user = persons_service.get_person_by_email(args["email"])
            if not user["active"]:
                return (
                    {"error": True, "message": "This user is inactive."},
                    400,
                )
        except PersonNotFoundException:
            return (
                {"error": True, "message": "Email not listed in database."},
                400,
            )

        token = auth_service.generate_reset_token()
        auth_tokens_store.add(
            "reset-token-%s" % args["email"], token, ttl=3600 * 2
        )
        params = {"email": args["email"], "token": token}
        query = urllib.parse.urlencode(params)
        reset_url = "%s://%s/reset-change-password?%s" % (
            config.DOMAIN_PROTOCOL,
            config.DOMAIN_NAME,
            query,
        )
        time_string = format_datetime(
            date_helpers.get_utc_now_datetime(),
            tzinfo=user["timezone"],
            locale=user["locale"],
        )
        person_IP = request.headers.get("X-Forwarded-For", None)
        organisation = persons_service.get_organisation()
        html = f"""<p>Hello {user["first_name"]},</p>

<p>
You have requested for a password reset. You can follow this link to change your
password: <a href="{reset_url}">{reset_url}</a>
</p>

<p>
This link will expire after 2 hours. After, you have to do a new request to reset your password.
This email was sent at this date: {time_string}.
The IP of the person who requested this is: {person_IP}.
</p>

Thank you and see you soon on Kitsu,
</p>
<p>
{organisation["name"]} Team
</p>
"""
        subject = f"{organisation['name']} - Kitsu: password recovery"
        emails.send_email(subject, html, args["email"])
        return {"success": "Reset token sent"}


class TOTPResource(Resource, ArgsMixin):
    """
    Resource to allow a user to enable/disable TOTP.
    """

    @jwt_required()
    @permissions.require_person
    def put(self):
        """
        Resource to allow a user to pre-enable TOTP.
        ---
        description: ""
        tags:
            - Authentication
        responses:
          200:
            description: TOTP enabled
          400:
            description: Invalid password
                         Wrong or expired token
                         Inactive user
        """
        try:
            totp_provisionning_uri, totp_secret = auth_service.pre_enable_totp(
                persons_service.get_current_user()["id"]
            )
            return {
                "totp_provisionning_uri": totp_provisionning_uri,
                "otp_secret": totp_secret,
            }
        except TOTPAlreadyEnabledException:
            return (
                {"error": True, "message": "TOTP already enabled."},
                400,
            )

    @jwt_required()
    @permissions.require_person
    def post(self):
        """
        Resource to allow a user to enable TOTP.
        ---
        description: ""
        tags:
            - Authentication
        responses:
          200:
            description: TOTP enabled
          400:
            description: Invalid password
                         Wrong or expired token
                         Inactive user
        """
        args = self.get_args([("totp", "", True)])

        try:
            otp_recovery_codes = auth_service.enable_totp(
                persons_service.get_current_user()["id"], args["totp"]
            )
            return {"otp_recovery_codes": otp_recovery_codes}
        except TOTPAlreadyEnabledException:
            return (
                {"error": True, "message": "TOTP already enabled."},
                400,
            )
        except WrongOTPException:
            return (
                {
                    "error": True,
                    "message": "TOTP verification failed.",
                    "wrong_OTP": True,
                },
                400,
            )

    @jwt_required()
    @permissions.require_person
    def delete(self):
        """
        Resource to allow a user to disable TOTP.
        ---
        description: ""
        tags:
            - Authentication
        responses:
          200:
            description: TOTP disabled
          400:
            description: TOTP not enabled
        """
        args = self.get_args(
            [
                ("totp", None, False),
                ("email_otp", None, False),
                ("fido_authentication_response", {}, False, dict),
                ("recovery_code", None, False),
            ]
        )

        try:
            person = persons_service.get_current_user(unsafe=True)
            if not auth_service.person_two_factor_authentication_enabled(
                person
            ):
                raise TOTPNotEnabledException
            if not auth_service.check_two_factor_authentication(
                person,
                args["totp"],
                args["email_otp"],
                args["fido_authentication_response"],
                args["recovery_code"],
            ):
                raise WrongOTPException
            auth_service.disable_totp(person["id"])
            return {"success": True}
        except TOTPNotEnabledException:
            return (
                {"error": True, "message": "TOTP not enabled."},
                400,
            )
        except (WrongOTPException, MissingOTPException):
            return (
                {
                    "error": True,
                    "message": "OTP verification failed.",
                    "wrong_OTP": True,
                },
                400,
            )


class EmailOTPResource(Resource, ArgsMixin):
    """
    Resource to allow a user to enable/disable OTP by email or to send an OTP
    by email.
    """

    def get(self):
        """
        Resource to send an OTP by email to user.
        ---
        description: ""
        tags:
            - Authentication
        responses:
          200:
            description: OTP by email sent
          400:
            description: Invalid password
                         Wrong or expired token
                         Inactive user
        """
        args = self.get_args(
            [
                ("email", None, True),
            ],
            location="values",
        )

        try:
            try:
                person = persons_service.get_person_by_email_desktop_login(
                    args["email"]
                )
            except PersonNotFoundException:
                raise WrongUserException()
            if not person["email_otp_enabled"]:
                raise EmailOTPNotEnabledException
            auth_service.send_email_otp(person)
            return {"success": True}
        except EmailOTPNotEnabledException:
            return (
                {"error": True, "message": "OTP by email not enabled."},
                400,
            )
        except WrongUserException:
            return (
                {"error": True, "message": "User not found."},
                404,
            )

    @jwt_required()
    @permissions.require_person
    def put(self):
        """
        Resource to allow a user to pre-enable OTP by email.
        ---
        description: ""
        tags:
            - Authentication
        responses:
          200:
            description: OTP by email enabled
          400:
            description: Invalid password
                         Wrong or expired token
                         Inactive user
        """
        try:
            auth_service.pre_enable_email_otp(
                persons_service.get_current_user()["id"]
            )
            return {"success": True}
        except EmailOTPAlreadyEnabledException:
            return (
                {"error": True, "message": "OTP by email already enabled."},
                400,
            )

    @jwt_required()
    @permissions.require_person
    def post(self):
        """
        Resource to allow a user to enable OTP by email.
        ---
        description: ""
        tags:
            - Authentication
        responses:
          200:
            description: OTP by email enabled
          400:
            description: Invalid password
                         Wrong or expired token
                         Inactive user
        """
        args = self.get_args([("email_otp", "", True)])

        try:
            otp_recovery_codes = auth_service.enable_email_otp(
                persons_service.get_current_user()["id"],
                args["email_otp"],
            )
            return {"otp_recovery_codes": otp_recovery_codes}
        except EmailOTPAlreadyEnabledException:
            return (
                {"error": True, "message": "OTP by email already enabled."},
                400,
            )
        except WrongOTPException:
            return (
                {
                    "error": True,
                    "message": "OTP by email verification failed.",
                    "wrong_OTP": True,
                },
                400,
            )

    @jwt_required()
    @permissions.require_person
    def delete(self):
        """
        Resource to allow a user to disable OTP by email.
        ---
        description: ""
        tags:
            - Authentication
        responses:
          200:
            description: OTP by email disabled.
          400:
            description: Invalid password.
                         Wrong or expired token.
                         Inactive user.
                         Wrong 2FA.
        """
        args = self.get_args(
            [
                ("totp", None, False),
                ("email_otp", None, False),
                ("fido_authentication_response", {}, False, dict),
                ("recovery_code", None, False),
            ]
        )

        try:
            person = persons_service.get_current_user(unsafe=True)
            if not auth_service.person_two_factor_authentication_enabled(
                person
            ):
                raise EmailOTPNotEnabledException
            if not auth_service.check_two_factor_authentication(
                person,
                args["totp"],
                args["email_otp"],
                args["fido_authentication_response"],
                args["recovery_code"],
            ):
                raise WrongOTPException
            auth_service.disable_email_otp(person["id"])
            return {"success": True}
        except EmailOTPNotEnabledException:
            return (
                {"error": True, "message": "OTP by email not enabled."},
                400,
            )
        except (WrongOTPException, MissingOTPException):
            return (
                {
                    "error": True,
                    "message": "OTP verification failed.",
                    "wrong_OTP": True,
                },
                400,
            )


class FIDOResource(Resource, ArgsMixin):
    """
    Resource to allow a user to register/unregister FIDO device or to get a
    challenge for a FIDO device.
    """

    def get(self):
        """
        Resource to get a challenge for a FIDO device.
        ---
        description: ""
        tags:
            - Authentication
        responses:
          200:
            description: Challenge for FIDO device.
          400:
            description: Wrong parameter.
        """
        args = self.get_args(
            [
                ("email", None, True),
            ],
            location="values",
        )

        try:
            try:
                person = persons_service.get_person_by_email_desktop_login(
                    args["email"]
                )
            except PersonNotFoundException:
                raise WrongUserException()
            if not person["fido_enabled"]:
                raise FIDONotEnabledException
            return auth_service.get_challenge_fido(person["id"])
        except FIDONotEnabledException:
            return (
                {"error": True, "message": "FIDO not enabled."},
                400,
            )
        except WrongUserException:
            return (
                {"error": True, "message": "User not found."},
                404,
            )

    @jwt_required()
    @permissions.require_person
    def put(self):
        """
        Resource to allow a user to pre-register a FIDO device.
        ---
        description: ""
        tags:
            - Authentication
        responses:
          200:
            description: FIDO device pre-registered.
          400:
            description: Invalid password
                         Wrong or expired token
                         Inactive user
        """
        return auth_service.pre_register_fido(
            persons_service.get_current_user()["id"]
        )

    @permissions.require_person
    @jwt_required()
    def post(self):
        """
        Resource to allow a user to register a FIDO device.
        ---
        description: ""
        tags:
            - Authentication
        responses:
          200:
            description: FIDO device registered.
          400:
            description: Invalid password
                         Wrong or expired token
                         Inactive user
        """
        try:
            args = self.get_args(
                [
                    ("registration_response", {}, True, dict),
                    ("device_name", "", True),
                ]
            )

            otp_recovery_codes = auth_service.register_fido(
                persons_service.get_current_user()["id"],
                args["registration_response"],
                args["device_name"],
            )
            return {"otp_recovery_codes": otp_recovery_codes}
        except FIDONoPreregistrationException:
            return (
                {"error": True, "message": "No preregistration before."},
                400,
            )
        except FIDOServerException:
            return (
                {
                    "error": True,
                    "message": "FIDO server exception your registration response is probly wrong.",
                },
                400,
            )

    @jwt_required()
    @permissions.require_person
    def delete(self):
        """
        Resource to allow a user to unregister a FIDO device.
        ---
        description: ""
        tags:
            - Authentication
        responses:
          200:
            description: FIDO device unregistered.
          400:
            description: Invalid password
                         Wrong or expired token
                         Inactive user
                         Wrong 2FA
        """
        args = self.get_args(
            [
                ("totp", None, False),
                ("email_otp", None, False),
                ("fido_authentication_response", {}, False, dict),
                ("recovery_code", None, False),
                ("device_name", None, True),
            ]
        )

        try:
            person = persons_service.get_current_user(unsafe=True)
            if not auth_service.person_two_factor_authentication_enabled(
                person
            ):
                raise FIDONotEnabledException
            if not auth_service.check_two_factor_authentication(
                person,
                args["totp"],
                args["email_otp"],
                args["fido_authentication_response"],
                args["recovery_code"],
            ):
                raise WrongOTPException
            auth_service.unregister_fido(person["id"], args["device_name"])
            return {"success": True}
        except FIDONotEnabledException:
            return (
                {"error": True, "message": "FIDO not enabled."},
                400,
            )
        except (WrongOTPException, MissingOTPException):
            return (
                {
                    "error": True,
                    "message": "OTP verification failed.",
                    "wrong_OTP": True,
                },
                400,
            )


class RecoveryCodesResource(Resource, ArgsMixin):
    """
    Resource to allow a user to generate new recovery codes.
    """

    @jwt_required()
    @permissions.require_person
    def put(self):
        """
        Resource to allow a user to generate new recovery codes.
        ---
        description: ""
        tags:
            - Authentication
        responses:
          200:
            description: new recovery codes.
          400:
            description: Invalid password
                         Wrong or expired token
                         Inactive user
        """
        args = self.get_args(
            [
                ("totp", None, False),
                ("email_otp", None, False),
                ("fido_authentication_response", {}, False, dict),
                ("recovery_code", None, False),
            ]
        )

        try:
            person = persons_service.get_current_user(unsafe=True)
            if not auth_service.person_two_factor_authentication_enabled(
                person
            ):
                raise NoTwoFactorAuthenticationEnabled
            if not auth_service.check_two_factor_authentication(
                person,
                args["totp"],
                args["email_otp"],
                args["fido_authentication_response"],
                args["recovery_code"],
            ):
                raise WrongOTPException
            otp_recovery_codes = auth_service.generate_new_recovery_codes(
                person["id"]
            )
            return {"otp_recovery_codes": otp_recovery_codes}
        except WrongOTPException:
            return (
                {
                    "error": True,
                    "message": "OTP verification failed.",
                    "wrong_OTP": True,
                },
                400,
            )
        except NoTwoFactorAuthenticationEnabled:
            return (
                {
                    "error": True,
                    "message": "No two factor authentication enabled.",
                },
                400,
            )


class SAMLSSOResource(Resource, ArgsMixin):
    """
    Resource to allow a user to login with SAML SSO.
    """

    def post(self):
        """
        Resource to allow a user to login with SAML SSO.
        ---
        description: ""
        tags:
            - Authentication
        responses:
          302:
            description: Login successful, redirect to the home page.
          400:
            description: Wrong parameter
        """
        if not config.SAML_ENABLED:
            return {"error": "SAML is not enabled."}, 400
        authn_response = current_app.extensions[
            "saml_client"
        ].parse_authn_request_response(
            request.form["SAMLResponse"], entity.BINDING_HTTP_POST
        )
        authn_response.get_identity()
        email = authn_response.get_subject().text
        person_info = {
            k: (
                " ".join(v)
                if isinstance(v, list) and k in ["first_name", "last_name"]
                else v
            )
            for k, v in authn_response.ava.items()
            if k
            in [
                "first_name",
                "last_name",
                "phone",
                "role",
                "departments",
                "studio_id",
                "active",
            ]
        }
        try:
            user = persons_service.get_person_by_email(email)
            for k, v in person_info.items():
                if user.get(k) != v:
                    persons_service.update_person(
                        user["id"], person_info, bypass_protected_accounts=True
                    )
                    break
        except PersonNotFoundException:
            user = persons_service.create_person(
                email, "default".encode("utf-8"), **person_info
            )

        response = make_response(
            redirect(f"{config.DOMAIN_PROTOCOL}://{config.DOMAIN_NAME}")
        )

        if user["active"]:
            access_token = create_access_token(
                identity=user["id"],
                additional_claims={
                    "identity_type": "person",
                },
            )
            refresh_token = create_refresh_token(
                identity=user["id"],
                additional_claims={
                    "identity_type": "person",
                },
            )
            identity_changed.send(
                current_app._get_current_object(),
                identity=Identity(user["id"], "person"),
            )

            ip_address = request.environ.get(
                "HTTP_X_REAL_IP", request.remote_addr
            )

            set_access_cookies(response, access_token)
            set_refresh_cookies(response, refresh_token)
            events_service.create_login_log(user["id"], ip_address, "web")

        return response


class SAMLLoginResource(Resource, ArgsMixin):
    """
    Resource to allow a user to login with SAML SSO.
    """

    def get(self):
        """
        Resource to allow a user to login with SAML SSO.
        ---
        description: ""
        tags:
            - Authentication
        responses:
          302:
            description: Redirect to the SAML IDP.
          400:
            description: Wrong parameter.
        """
        if not config.SAML_ENABLED:
            return {"error": "SAML is not enabled."}, 400

        try:
            _, info = current_app.extensions[
                "saml_client"
            ].prepare_for_authenticate()
        except client_base.SAMLError:
            # retry with new client
            current_app.extensions["saml_client"] = saml_client_for(
                config.SAML_METADATA_URL
            )
            _, info = current_app.extensions[
                "saml_client"
            ].prepare_for_authenticate()

        redirect_url = None

        # Select the IdP URL to send the AuthN request to
        for key, value in info["headers"]:
            if key == "Location":
                redirect_url = value

        return redirect(redirect_url, code=302)



"""Add Comment.updated_by

Revision ID: 06552e22f9e7
Revises: fba149993140
Create Date: 2025-02-27 13:44:58.067787

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "06552e22f9e7"
down_revision = "fba149993140"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("comment", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "updated_by",
                sqlalchemy_utils.types.uuid.UUIDType(binary=False),
                default=uuid.uuid4,
                nullable=True,
            )
        )
        batch_op.create_index(
            batch_op.f("ix_comment_updated_by"), ["updated_by"], unique=False
        )
        batch_op.create_foreign_key(
            "comment_updated_by_fkey", "person", ["updated_by"], ["id"]
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("comment", schema=None) as batch_op:
        batch_op.drop_constraint("comment_updated_by_fkey", type_="foreignkey")
        batch_op.drop_index(batch_op.f("ix_comment_updated_by"))
        batch_op.drop_column("updated_by")

    # ### end Alembic commands ###



from zou.app import config

REDIS_HOST = config.KEY_VALUE_STORE["host"]
REDIS_PORT = config.KEY_VALUE_STORE["port"]
REDIS_DB = config.KV_JOB_DB_INDEX
REDIS_PASSWORD = config.KEY_VALUE_STORE["password"]



from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin
from sqlalchemy.sql import expression


class TaskStatus(db.Model, BaseMixin, SerializerMixin):
    """
    Describe the state of a task. A status marked as reviewable expects a
    preview file linked to relate comment.
    """

    name = db.Column(db.String(40), nullable=False)
    archived = db.Column(db.Boolean(), default=False)
    short_name = db.Column(
        db.String(10), unique=True, nullable=False, index=True
    )
    description = db.Column(db.Text())
    color = db.Column(db.String(7), nullable=False)
    priority = db.Column(db.Integer, default=1)

    is_done = db.Column(db.Boolean(), default=False, index=True)
    is_artist_allowed = db.Column(db.Boolean(), default=True)
    is_client_allowed = db.Column(db.Boolean(), default=True)
    is_retake = db.Column(db.Boolean(), default=False)
    is_feedback_request = db.Column(db.Boolean(), default=False, index=True)
    is_default = db.Column(db.Boolean(), default=False, index=True)
    is_wip = db.Column(
        db.Boolean(),
        server_default=expression.false(),
        default=False,
        index=True,
    )
    shotgun_id = db.Column(db.Integer)

    for_concept = db.Column(
        db.Boolean(), server_default=expression.false(), default=False
    )



import sys

from zou.app.blueprints.assets import blueprint as assets_blueprint
from zou.app.blueprints.auth import blueprint as auth_blueprint
from zou.app.blueprints.breakdown import blueprint as breakdown_blueprint
from zou.app.blueprints.chats import blueprint as chats_blueprint
from zou.app.blueprints.comments import blueprint as comments_blueprint
from zou.app.blueprints.crud import blueprint as crud_blueprint
from zou.app.blueprints.departments import blueprint as departments_blueprint
from zou.app.blueprints.entities import blueprint as entities_blueprint
from zou.app.blueprints.events import blueprint as events_blueprint
from zou.app.blueprints.export import blueprint as export_blueprint
from zou.app.blueprints.files import blueprint as files_blueprint
from zou.app.blueprints.index import blueprint as index_blueprint
from zou.app.blueprints.search import blueprint as search_blueprint
from zou.app.blueprints.news import blueprint as news_blueprint
from zou.app.blueprints.persons import blueprint as persons_blueprint
from zou.app.blueprints.playlists import blueprint as playlists_blueprint
from zou.app.blueprints.projects import blueprint as projects_blueprint
from zou.app.blueprints.previews import blueprint as previews_blueprint
from zou.app.blueprints.source import blueprint as import_blueprint
from zou.app.blueprints.shots import blueprint as shots_blueprint
from zou.app.blueprints.tasks import blueprint as tasks_blueprint
from zou.app.blueprints.user import blueprint as user_blueprint
from zou.app.blueprints.edits import blueprint as edits_blueprint
from zou.app.blueprints.concepts import blueprint as concepts_blueprint

from zou.app.utils.plugins import load_plugins
from zou.app.utils import events


def configure(app):
    """
    Turn Flask app into a REST API. It configures routes, auth and events
    system.
    """
    app.url_map.strict_slashes = False
    configure_api_routes(app)
    register_event_handlers(app)
    load_plugins(app)
    return app


def configure_api_routes(app):
    """
    Register blueprints (modules). Each blueprint describe routes and
    associated resources (controllers).
    """
    app.register_blueprint(auth_blueprint)
    app.register_blueprint(assets_blueprint)
    app.register_blueprint(breakdown_blueprint)
    app.register_blueprint(chats_blueprint)
    app.register_blueprint(comments_blueprint)
    app.register_blueprint(crud_blueprint)
    app.register_blueprint(departments_blueprint)
    app.register_blueprint(entities_blueprint)
    app.register_blueprint(export_blueprint)
    app.register_blueprint(events_blueprint)
    app.register_blueprint(files_blueprint)
    app.register_blueprint(import_blueprint)
    app.register_blueprint(index_blueprint)
    app.register_blueprint(news_blueprint)
    app.register_blueprint(persons_blueprint)
    app.register_blueprint(playlists_blueprint)
    app.register_blueprint(projects_blueprint)
    app.register_blueprint(shots_blueprint)
    app.register_blueprint(tasks_blueprint)
    app.register_blueprint(previews_blueprint)
    app.register_blueprint(user_blueprint)
    app.register_blueprint(edits_blueprint)
    app.register_blueprint(search_blueprint)
    app.register_blueprint(concepts_blueprint)
    return app


def register_event_handlers(app):
    """
    Load code from event handlers folder. Then it registers in the event manager
    each event handler listed in the __init_.py.
    """
    sys.path.insert(0, app.config["EVENT_HANDLERS_FOLDER"])
    try:
        import event_handlers

        events.register_all(event_handlers.event_map, app)
    except ImportError:
        # Event handlers folder is not properly configured.
        # Handlers are optional, that's why this error is ignored.
        # app.logger.info("No event handlers folder is configured.")
        pass
    return app



"""Change person.role to ChoiceType

Revision ID: 5798d2c9020b
Revises: 693cc511d28d
Create Date: 2024-01-12 16:34:19.194350

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils


# revision identifiers, used by Alembic.
revision = "5798d2c9020b"
down_revision = "693cc511d28d"
branch_labels = None
depends_on = None

ROLE_TYPES = [
    ("user", "Artist"),
    ("admin", "Studio Manager"),
    ("supervisor", "Supervisor"),
    ("manager", "Production Manager"),
    ("client", "Client"),
    ("vendor", "Vendor"),
]


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("person", schema=None) as batch_op:
        batch_op.alter_column(
            "role",
            existing_type=sa.VARCHAR(length=30),
            type_=sqlalchemy_utils.types.choice.ChoiceType(ROLE_TYPES),
            existing_nullable=True,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("person", schema=None) as batch_op:
        batch_op.alter_column(
            "role",
            existing_type=sqlalchemy_utils.types.choice.ChoiceType(ROLE_TYPES),
            type_=sa.VARCHAR(length=30),
            existing_nullable=True,
        )

    # ### end Alembic commands ###



from flask import current_app
from flask_restful import Resource

from zou.app.models.department import Department
from zou.app.models.task_type import TaskType
from zou.app.utils import colors
from zou.app.services import tasks_service

from zou.app.blueprints.source.shotgun.base import (
    BaseImportShotgunResource,
    ImportRemoveShotgunBaseResource,
)


class ImportShotgunStepsResource(BaseImportShotgunResource):
    def __init__(self):
        Resource.__init__(self)

    def extract_data(self, sg_step):
        color = self.extract_color(sg_step)
        department_name = self.extract_department_name(sg_step)
        return {
            "name": sg_step["code"],
            "short_name": sg_step.get("short_name", ""),
            "shotgun_id": sg_step["id"],
            "color": color,
            "department_name": department_name,
            "for_entity": sg_step.get("entity_type", "Asset"),
        }

    def extract_color(self, sg_step):
        color = sg_step.get("color", "0,0,0")
        return colors.rgb_to_hex(color)

    def extract_department_name(self, sg_step):
        splitted_name = sg_step["code"].split(" ")
        department_name = splitted_name[0]
        return department_name

    def import_entry(self, data):
        department = self.save_department(data)
        return self.save_task_type(department, data)

    def save_department(self, data):
        department = Department.get_by(name=data["department_name"])
        if department is None:
            department_data = {
                "name": data["department_name"],
                "color": data["color"],
            }
            department = Department(**department_data)
            department.save()
            current_app.logger.info("Department created: %s" % department)
        del data["department_name"]
        return department

    def save_task_type(self, department, data):
        task_type = TaskType.get_by(shotgun_id=data["shotgun_id"])
        data["department_id"] = department.id
        if task_type is None:
            task_type = TaskType.get_by(
                name=data["name"], for_entity=data["for_entity"]
            )

        if task_type is None:
            task_type = TaskType(**data)
            task_type.save()
            current_app.logger.info("Task Type created: %s" % task_type)
        else:
            existing_task_type = TaskType.get_by(
                name=data["name"],
                for_entity=data["for_entity"],
                department_id=data["department_id"],
            )
            if existing_task_type is not None:
                data.pop("name", None)
                data.pop("for_entity", None)
                data.pop("department_id", None)
            task_type.update(data)
            tasks_service.clear_task_type_cache(str(task_type.id))
            current_app.logger.info("Task Type updated: %s" % task_type)
        return task_type


class ImportRemoveShotgunStepResource(ImportRemoveShotgunBaseResource):
    def __init__(self):
        ImportRemoveShotgunBaseResource.__init__(self, TaskType)



"""add departement links tables for hardware and software

Revision ID: dde6be40f54f
Revises: 9af2df17a9d5
Create Date: 2025-07-16 21:58:45.737760

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "dde6be40f54f"
down_revision = "9af2df17a9d5"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "hardware_item_department_link",
        sa.Column(
            "department_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column(
            "hardware_item_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column(
            "id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["department_id"],
            ["department.id"],
        ),
        sa.ForeignKeyConstraint(
            ["hardware_item_id"],
            ["hardware_item.id"],
        ),
        sa.PrimaryKeyConstraint("department_id", "hardware_item_id", "id"),
        sa.UniqueConstraint(
            "hardware_item_id",
            "department_id",
            name="hardware_item_department_link_uc",
        ),
    )
    with op.batch_alter_table(
        "hardware_item_department_link", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_hardware_item_department_link_department_id"),
            ["department_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_hardware_item_department_link_hardware_item_id"),
            ["hardware_item_id"],
            unique=False,
        )

    op.create_table(
        "software_department_link",
        sa.Column(
            "department_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column(
            "software_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column(
            "id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["department_id"],
            ["department.id"],
        ),
        sa.ForeignKeyConstraint(
            ["software_id"],
            ["software.id"],
        ),
        sa.PrimaryKeyConstraint("department_id", "software_id", "id"),
        sa.UniqueConstraint(
            "department_id", "software_id", name="software_department_link_uc"
        ),
    )
    with op.batch_alter_table(
        "software_department_link", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_software_department_link_department_id"),
            ["department_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_software_department_link_software_id"),
            ["software_id"],
            unique=False,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "software_department_link", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_software_department_link_software_id")
        )
        batch_op.drop_index(
            batch_op.f("ix_software_department_link_department_id")
        )

    op.drop_table("software_department_link")
    with op.batch_alter_table(
        "hardware_item_department_link", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_hardware_item_department_link_hardware_item_id")
        )
        batch_op.drop_index(
            batch_op.f("ix_hardware_item_department_link_department_id")
        )

    op.drop_table("hardware_item_department_link")
    # ### end Alembic commands ###



"""Add totp columns for Person

Revision ID: 8a1b4a1b7f4a
Revises: a6c25eed3ea1
Create Date: 2022-11-15 23:15:42.600126

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "8a1b4a1b7f4a"
down_revision = "a6c25eed3ea1"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "person", sa.Column("totp_enabled", sa.Boolean(), nullable=True)
    )
    op.add_column(
        "person", sa.Column("totp_secret", sa.String(length=32), nullable=True)
    )
    op.add_column(
        "person",
        sa.Column(
            "otp_recovery_codes",
            sa.ARRAY(sa.LargeBinary(length=60)),
            nullable=True,
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("person", "otp_recovery_codes")
    op.drop_column("person", "totp_secret")
    op.drop_column("person", "totp_enabled")
    # ### end Alembic commands ###



import datetime

from dateutil import relativedelta

from sqlalchemy import func
from sqlalchemy.exc import DataError
from sqlalchemy.orm import aliased

from zou.app.models.day_off import DayOff
from zou.app.models.department import Department
from zou.app.models.project import Project
from zou.app.models.task import Task
from zou.app.models.task_type import TaskType
from zou.app.models.time_spent import TimeSpent
from zou.app.models.entity import Entity
from zou.app.models.entity_type import EntityType
from zou.app.models.person import Person

from zou.app.utils import fields, date_helpers

from zou.app.services import user_service
from zou.app.services.exception import WrongDateFormatException


def get_time_spents_for_entity(entity_id):
    """
    Return all time spents related to given entity.
    """
    query = (
        TimeSpent.query.join(Task)
        .filter(Task.entity_id == entity_id)
        .order_by(TimeSpent.date.desc())
    )
    return TimeSpent.serialize_list(query.all())


def get_year_table(
    person_id=None, project_id=None, department_ids=None, studio_id=None
):
    """
    Return a table giving time spent by user and by month for given year.
    """
    return get_yearly_table(
        None,
        detail_level="year",
        person_id=person_id,
        project_id=project_id,
        department_ids=department_ids,
        studio_id=studio_id,
    )


def get_month_table(
    year, person_id=None, project_id=None, department_ids=None, studio_id=None
):
    """
    Return a table giving time spent by user and by month for given year.
    """
    return get_yearly_table(
        year,
        person_id=person_id,
        project_id=project_id,
        department_ids=department_ids,
        studio_id=studio_id,
    )


def get_week_table(
    year, person_id=None, project_id=None, department_ids=None, studio_id=None
):
    """
    Return a table giving time spent by user and by week for given year.
    """
    return get_yearly_table(
        year,
        "week",
        person_id=person_id,
        project_id=project_id,
        department_ids=department_ids,
        studio_id=studio_id,
    )


def get_day_table(
    year,
    month,
    person_id=None,
    project_id=None,
    department_ids=None,
    studio_id=None,
):
    """
    Return a table giving time spent by user and by day for given year and
    month.
    """
    time_spents = get_time_spents_for_month(
        year,
        month,
        person_id=person_id,
        project_id=project_id,
        department_ids=department_ids,
        studio_id=studio_id,
    )
    return get_table_from_time_spents(time_spents, "day")


def get_yearly_table(
    year=None,
    detail_level="month",
    person_id=None,
    project_id=None,
    department_ids=None,
    studio_id=None,
):
    """
    Return a table giving time spent by user and by week or month for given
    year. Week or month detail level can be selected through *detail_level*
    argument.
    """
    time_spents = get_time_spents_for_year(
        year=year,
        person_id=person_id,
        project_id=project_id,
        department_ids=department_ids,
        studio_id=studio_id,
    )
    return get_table_from_time_spents(time_spents, detail_level)


def get_time_spents_for_year(
    year=None,
    person_id=None,
    project_id=None,
    department_ids=None,
    studio_id=None,
):
    """
    Return all time spents for given year.
    """
    query = TimeSpent.query

    if person_id is not None:
        query = query.filter(TimeSpent.person_id == person_id)

    if year is not None:
        query = query.filter(
            TimeSpent.date.between(
                datetime.datetime(int(year), 1, 1),
                datetime.datetime(int(year), 12, 31),
            )
        )

    if project_id is not None or department_ids is not None:
        query = query.join(Task)

    if project_id is not None:
        if isinstance(project_id, list):
            query = query.filter(Task.project_id.in_(project_id))
        else:
            query = query.filter(Task.project_id == project_id)

    if department_ids:
        query = query.join(TaskType).filter(
            TaskType.department_id.in_(department_ids)
        )

    if studio_id is not None:
        query = query.join(Person, Person.id == TimeSpent.person_id).filter(
            Person.studio_id == studio_id
        )

    return query.all()


def get_time_spents_for_month(
    year,
    month,
    person_id=None,
    project_id=None,
    department_ids=None,
    studio_id=None,
):
    """
    Return all time spents for given month.
    """
    date = datetime.datetime(int(year), int(month), 1)
    next_month = date + relativedelta.relativedelta(months=1)
    query = TimeSpent.query.filter(TimeSpent.date >= date).filter(
        TimeSpent.date < next_month
    )

    if person_id is not None:
        query = query.filter(TimeSpent.person_id == person_id)

    if project_id is not None or department_ids is not None:
        query = query.join(Task)

    if project_id is not None:
        if isinstance(project_id, list):
            query = query.filter(Task.project_id.in_(project_id))
        else:
            query = query.filter(Task.project_id == project_id)

    if department_ids:
        query = query.join(TaskType).filter(
            TaskType.department_id.in_(department_ids)
        )

    if studio_id is not None:
        query = query.join(Person, Person.id == TimeSpent.person_id).filter(
            Person.studio_id == studio_id
        )

    return query.all()


def get_table_from_time_spents(time_spents, detail_level="month"):
    """
    Buid a time spent table based on given time spents and given level
    of detail (week, day or month).
    """
    result = {}
    for time_spent in time_spents:
        if detail_level == "week":
            unit = str(time_spent.date.isocalendar()[1])
        elif detail_level == "day":
            unit = str(time_spent.date.day)
        elif detail_level == "year":
            unit = str(time_spent.date.year)
        else:
            unit = str(time_spent.date.month)

        person_id = str(time_spent.person_id)
        if unit not in result:
            result[unit] = {}
        if person_id not in result[unit]:
            result[unit][person_id] = 0
        result[unit][person_id] += time_spent.duration
    return result


def get_time_spents(
    person_id, date=None, project_ids=None, department_ids=None
):
    """
    Return time spents for given person and date.
    """
    try:
        query = TimeSpent.query.filter_by(
            person_id=person_id, date=func.cast(date, TimeSpent.date.type)
        )

        if project_ids is not None or department_ids is not None:
            query = query.join(Task)

        if project_ids is not None:
            query = query.filter(Task.project_id.in_(project_ids))

        if department_ids:
            query = query.join(TaskType)
            query = query.filter(TaskType.department_id.in_(department_ids))
    except DataError:
        raise WrongDateFormatException
    return fields.serialize_list(query.all())


def get_time_spents_range(person_id, start_date, end_date):
    """
    Returns time spents for person and date range.
    """
    try:
        query = TimeSpent.query.filter_by(person_id=person_id)
        time_spents = query.filter(
            TimeSpent.date.between(
                func.cast(start_date, TimeSpent.date.type),
                func.cast(end_date, TimeSpent.date.type),
            )
        ).all()
    except DataError:
        raise WrongDateFormatException
    return fields.serialize_list(time_spents)


def get_time_spent(person_id, task_id, date):
    """
    Return time spent for given person, task and date.
    """
    try:
        time_spent = TimeSpent.query.filter_by(
            person_id=person_id,
            task_id=task_id,
            date=func.cast(date, TimeSpent.date.type),
        ).first()
    except DataError:
        raise WrongDateFormatException
    if time_spent is not None:
        return time_spent.serialize()
    else:
        return None


def get_day_off(person_id, date):
    """
    Return day off for given person and date.
    """
    try:
        day_off = DayOff.get_by(
            func.cast(date, DayOff.date.type).between(
                DayOff.date, DayOff.end_date
            ),
            person_id=person_id,
        )
    except DataError:
        raise WrongDateFormatException
    if day_off is not None:
        return day_off.serialize()
    else:
        return {}


def get_year_time_spents(
    person_id, year, project_id=None, department_ids=None
):
    """
    Return aggregated time spents at task level for given person and month.
    """
    start, end = date_helpers.get_year_interval(year)
    entries = get_person_time_spent_entries(
        person_id,
        TimeSpent.date >= func.cast(start, TimeSpent.date.type),
        TimeSpent.date < func.cast(end, TimeSpent.date.type),
        project_id=project_id,
        department_ids=department_ids,
    )
    return build_results(entries)


def get_month_time_spents(
    person_id, year, month, project_id=None, department_ids=None
):
    """
    Return aggregated time spents at task level for given person and month.
    """
    start, end = date_helpers.get_month_interval(year, month)
    entries = get_person_time_spent_entries(
        person_id,
        TimeSpent.date >= func.cast(start, TimeSpent.date.type),
        TimeSpent.date < func.cast(end, TimeSpent.date.type),
        project_id=project_id,
        department_ids=department_ids,
    )
    return build_results(entries)


def get_week_time_spents(
    person_id, year, week, project_id=None, department_ids=None
):
    """
    Return aggregated time spents at task level for given person and week.
    """
    start, end = date_helpers.get_week_interval(year, week)
    entries = get_person_time_spent_entries(
        person_id,
        TimeSpent.date >= func.cast(start, TimeSpent.date.type),
        TimeSpent.date < func.cast(end, TimeSpent.date.type),
        project_id=project_id,
        department_ids=department_ids,
    )
    return build_results(entries)


def get_day_time_spents(
    person_id, year, month, day, project_id=None, department_ids=None
):
    """
    Return aggregated time spents at task level for given person and day.
    """
    start, end = date_helpers.get_day_interval(year, month, day)
    entries = get_person_time_spent_entries(
        person_id,
        TimeSpent.date >= func.cast(start, TimeSpent.date.type),
        TimeSpent.date < func.cast(end, TimeSpent.date.type),
        project_id=project_id,
        department_ids=department_ids,
    )
    return build_results(entries)


def get_person_time_spent_entries(
    person_id, *args, project_id=None, department_ids=None
):
    """
    Return aggregated time spents at task level for given person and
    query filter (args).
    """
    Sequence = aliased(Entity, name="sequence")
    Episode = aliased(Entity, name="episode")
    query = (
        Task.query.with_entities(Task.id, Task.task_type_id)
        .join(Entity, Entity.id == Task.entity_id)
        .join(Project, Project.id == Task.project_id)
        .join(TimeSpent)
        .join(EntityType)
        .group_by(
            Task.id,
            Task.task_type_id,
            Project.id,
            Project.name,
            Entity.name,
            EntityType.name,
            Sequence.name,
            Episode.name,
        )
        .outerjoin(Sequence, Sequence.id == Entity.parent_id)
        .outerjoin(Episode, Episode.id == Sequence.parent_id)
        .filter(TimeSpent.person_id == person_id)
        .add_columns(
            Project.id,
            Project.name,
            Entity.name,
            EntityType.name,
            Sequence.name,
            Episode.name,
            func.sum(TimeSpent.duration),
        )
    )

    if project_id is not None:
        if isinstance(project_id, list):
            query = query.filter(Task.project_id.in_(project_id))
        else:
            query = query.filter(Task.project_id == project_id)

    if department_ids:
        query = query.join(TaskType, TaskType.id == Task.task_type_id).filter(
            TaskType.department_id.in_(department_ids)
        )

    for arg in args:
        query = query.filter(arg)

    return query.all()


def build_results(entries):
    """
    Build results with information to build a time sheet based on given entries
    (tasks + time spent duration aggregate)
    """
    result = []
    for (
        task_id,
        task_type_id,
        project_id,
        project_name,
        entity_name,
        entity_type_name,
        sequence_name,
        episode_name,
        duration,
    ) in entries:
        result.append(
            {
                "task_id": str(task_id),
                "task_type_id": str(task_type_id),
                "project_id": str(project_id),
                "project_name": project_name,
                "entity_name": entity_name,
                "entity_type_name": entity_type_name,
                "sequence_name": sequence_name,
                "episode_name": episode_name,
                "duration": duration,
            }
        )
    return result


def get_day_offs_for_month(year, month):
    """
    Get all day off entries for given year and month.
    """
    start, end = date_helpers.get_month_interval(year, month)
    return get_day_offs_between(start, end)


def get_person_day_offs_for_week(person_id, year, week):
    """
    Get all day off entries for given person, year and week.
    """
    start, end = date_helpers.get_week_interval(year, week)
    return get_day_offs_between(start, end, person_id=person_id)


def get_person_day_offs_for_month(person_id, year, month):
    """
    Get all day off entries for given person, year and week.
    """
    start, end = date_helpers.get_month_interval(year, month)
    return get_day_offs_between(start, end, person_id=person_id)


def get_person_day_offs_for_year(person_id, year):
    """
    Get all day off entries for given person, year.
    """
    start, end = date_helpers.get_year_interval(year)
    return get_day_offs_between(start, end, person_id=person_id)


def get_day_offs_between(
    start=None, end=None, person_id=None, exclude_id=None
):
    """
    Get all day off entries for given person, start and end date.
    """
    query = DayOff.query
    if person_id is not None:
        query = query.filter(DayOff.person_id == person_id)

    if start is not None:
        query = query.filter(
            func.cast(start, DayOff.end_date.type) <= DayOff.end_date
        )
    if end is not None:
        query = query.filter(func.cast(end, DayOff.date.type) >= DayOff.date)

    if exclude_id is not None:
        query = query.filter(DayOff.id != exclude_id)

    return DayOff.serialize_list(query.all())


def get_timezoned_interval(start, end):
    """
    Get time intervals adapted to the user timezone.
    """
    timezone = user_service.get_timezone()
    return date_helpers.get_timezoned_interval(start, end, timezone)


def get_project_month_time_spents(project_id, timezone=None):
    """
    Get aggregated time spents by department by person by month for given
    project.
    """
    data = {}
    query = (
        TimeSpent.query.join(Task)
        .join(TaskType, TaskType.id == Task.task_type_id)
        .join(Department, Department.id == TaskType.department_id)
        .filter(Task.project_id == project_id)
        .add_columns(Department.id)
        .order_by(TimeSpent.date)
    )

    for time_spent, department_id in query.all():
        date_key = date_helpers.get_simple_string_with_timezone_from_date(
            time_spent.date, timezone
        )[0:7]
        if department_id not in data:
            data[department_id] = {"total": 0}
        if time_spent.person_id not in data[department_id]:
            data[department_id][time_spent.person_id] = {"total": 0}
        if date_key not in data[department_id][time_spent.person_id]:
            data[department_id][time_spent.person_id][date_key] = 0

        data[department_id][time_spent.person_id][
            date_key
        ] += time_spent.duration
        data[department_id]["total"] += time_spent.duration
        data[department_id][time_spent.person_id][
            "total"
        ] += time_spent.duration
    return data



import semver
from pathlib import Path

from zou.app import config, db
from zou.app.models.plugin import Plugin
from zou.app.utils.plugins import (
    PluginManifest,
    run_plugin_migrations,
    downgrade_plugin_migrations,
    uninstall_plugin_files,
    install_plugin_files,
)


def install_plugin(path, force=False):
    """
    Install a plugin.
    """
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"Plugin path '{path}' does not exist.")

    manifest = PluginManifest.from_plugin_path(path)
    plugin = Plugin.query.filter_by(plugin_id=manifest.id).one_or_none()

    try:
        if plugin:
            current = semver.Version.parse(plugin.version)
            new = semver.Version.parse(str(manifest.version))
            if not force and new <= current:
                raise ValueError(
                    f"Plugin version {new} is not newer than {current}."
                )
            plugin.update_no_commit(manifest.to_model_dict())
        else:
            plugin = Plugin.create_no_commit(**manifest.to_model_dict())

        plugin_path = install_plugin_files(
            path, Path(config.PLUGIN_FOLDER) / manifest.id
        )
        run_plugin_migrations(plugin_path, plugin)
    except Exception:
        uninstall_plugin_files(manifest.id)
        db.session.rollback()
        db.session.remove()
        raise

    Plugin.commit()
    return plugin.serialize()


def uninstall_plugin(plugin_id):
    """
    Uninstall a plugin.
    """
    plugin_path = Path(config.PLUGIN_FOLDER) / plugin_id
    downgrade_plugin_migrations(plugin_path)
    installed = uninstall_plugin_files(plugin_path)
    plugin = Plugin.query.filter_by(plugin_id=plugin_id).one_or_none()
    if plugin is not None:
        installed = True
        plugin.delete()

    if not installed:
        raise ValueError(f"Plugin '{plugin_id}' is not installed.")
    return True



"""Add new table ProductionScheduleVersion

Revision ID: 5f715f2b6348
Revises: dde6be40f54f
Create Date: 2025-07-11 03:57:29.859765

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "5f715f2b6348"
down_revision = "dde6be40f54f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "production_schedule_version",
        sa.Column("name", sa.String(length=80), nullable=False),
        sa.Column(
            "project_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column(
            "production_schedule_from",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=True,
        ),
        sa.Column("locked", sa.Boolean(), nullable=True),
        sa.Column("canceled", sa.Boolean(), nullable=True),
        sa.Column(
            "id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["production_schedule_from"],
            ["production_schedule_version.id"],
        ),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table(
        "production_schedule_version", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f(
                "ix_production_schedule_version_production_schedule_from"
            ),
            ["production_schedule_from"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_production_schedule_version_project_id"),
            ["project_id"],
            unique=False,
        )

    op.create_table(
        "production_schedule_version_task_link",
        sa.Column(
            "production_schedule_version_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column(
            "task_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column("start_date", sa.DateTime(), nullable=True),
        sa.Column("due_date", sa.DateTime(), nullable=True),
        sa.Column(
            "id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["production_schedule_version_id"],
            ["production_schedule_version.id"],
        ),
        sa.ForeignKeyConstraint(
            ["task_id"],
            ["task.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("production_schedule_version_task_link")
    with op.batch_alter_table(
        "production_schedule_version", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_production_schedule_version_project_id")
        )
        batch_op.drop_index(
            batch_op.f(
                "ix_production_schedule_version_production_schedule_from"
            )
        )

    op.drop_table("production_schedule_version")
    # ### end Alembic commands ###



"""empty message

Revision ID: 0ef6416a507b
Revises: 3fee3bd10f9d
Create Date: 2018-10-31 17:26:38.172799

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "0ef6416a507b"
down_revision = "3fee3bd10f9d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "asset_instance",
        sa.Column(
            "target_asset_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=True,
        ),
    )
    op.create_index(
        op.f("ix_asset_instance_target_asset_id"),
        "asset_instance",
        ["target_asset_id"],
        unique=False,
    )
    op.drop_constraint("asset_instance_uc", "asset_instance", type_="unique")
    op.create_unique_constraint(
        "asset_instance_uc",
        "asset_instance",
        ["asset_id", "target_asset_id", "scene_id", "number"],
    )
    op.drop_constraint(
        "asset_instance_name_uc", "asset_instance", type_="unique"
    )
    op.create_foreign_key(
        None, "asset_instance", "entity", ["target_asset_id"], ["id"]
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "asset_instance", type_="foreignkey")
    op.create_unique_constraint(
        "asset_instance_name_uc", "asset_instance", ["scene_id", "name"]
    )
    op.drop_constraint("asset_instance_uc", "asset_instance", type_="unique")
    op.create_unique_constraint(
        "asset_instance_uc",
        "asset_instance",
        ["asset_id", "scene_id", "number"],
    )
    op.drop_index(
        op.f("ix_asset_instance_target_asset_id"), table_name="asset_instance"
    )
    op.drop_column("asset_instance", "target_asset_id")
    # ### end Alembic commands ###



"""Add productionversionschedule uc

Revision ID: 26f96f65cfa3
Revises: 3d8e68dffeee
Create Date: 2025-07-18 11:49:59.302576

"""

from alembic import op


# revision identifiers, used by Alembic.
revision = "26f96f65cfa3"
down_revision = "3d8e68dffeee"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "production_schedule_version", schema=None
    ) as batch_op:
        batch_op.create_unique_constraint(
            "production_schedule_version_uc", ["name", "project_id"]
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "production_schedule_version", schema=None
    ) as batch_op:
        batch_op.drop_constraint(
            "production_schedule_version_uc", type_="unique"
        )

    # ### end Alembic commands ###



# coding: utf-8

import os
import datetime
import tempfile
import sys
import shutil
import click
import orjson as json

from tabulate import tabulate
from ldap3 import Server, Connection, ALL, NTLM, SIMPLE
from zou.app.utils import thumbnail as thumbnail_utils, auth
from zou.app.stores import auth_tokens_store, file_store, queue_store
from zou.app.services import (
    assets_service,
    backup_service,
    breakdown_service,
    deletion_service,
    edits_service,
    index_service,
    persons_service,
    preview_files_service,
    projects_service,
    shots_service,
    sync_service,
    tasks_service,
)
from zou.app.models.person import Person
from zou.app.models.preview_file import PreviewFile
from zou.app.models.task import Task
from zou.app.models.plugin import Plugin
from sqlalchemy.sql.expression import not_

from zou.app.services.exception import (
    PersonNotFoundException,
    IsUserLimitReachedException,
)

from zou.app import config

from zou.app import app


def clean_auth_tokens():
    """
    Remove all revoked tokens from the key value
    store.
    """
    for key in auth_tokens_store.keys():
        if auth_tokens_store.is_revoked(key):
            auth_tokens_store.delete(key)


def clear_all_auth_tokens():
    """
    Remove all authentication tokens from the key value store.
    """
    for key in auth_tokens_store.keys():
        auth_tokens_store.delete(key)


def init_data():
    """
    Put the minimum required data into the database to start with it.
    """
    with app.app_context():
        projects_service.get_open_status()
        projects_service.get_closed_status()
        print("Project status initialized.")

        assets_service.get_or_create_asset_type("Character")
        assets_service.get_or_create_asset_type("Prop")
        assets_service.get_or_create_asset_type("Environment")
        assets_service.get_or_create_asset_type("FX")
        print("Asset types initialized.")

        shots_service.get_episode_type()
        shots_service.get_sequence_type()
        shots_service.get_shot_type()
        print("Shot types initialized.")

        edits_service.get_edit_type()
        print("Edit type initialized.")

        modeling = tasks_service.get_or_create_department(
            "Modeling", "#78909C"
        )
        animation = tasks_service.get_or_create_department(
            "Animation", "#009688"
        )
        fx = tasks_service.get_or_create_department("FX", "#26C6DA")
        compositing = tasks_service.get_or_create_department(
            "Compositing", "#F06292"
        )
        concept = tasks_service.get_or_create_department("Concept", "#8D6E63")
        layout = tasks_service.get_or_create_department("Layout", "#7CB342")

        tasks_service.get_or_create_task_type(concept, "Concept", "#8D6E63", 1)
        tasks_service.get_or_create_task_type(
            modeling, "Modeling", "#78909C", 2
        )
        tasks_service.get_or_create_task_type(
            modeling, "Shading", "#64B5F6", 3
        )
        tasks_service.get_or_create_task_type(
            animation, "Rigging", "#9CCC65", 4
        )

        tasks_service.get_or_create_task_type(
            concept,
            "Storyboard",
            "#43A047",
            priority=1,
            for_entity="Shot",
        )
        tasks_service.get_or_create_task_type(
            layout,
            "Layout",
            "#7CB342",
            priority=2,
            for_entity="Shot",
        )
        tasks_service.get_or_create_task_type(
            animation,
            "Animation",
            "#009688",
            priority=3,
            for_entity="Shot",
        )
        tasks_service.get_or_create_task_type(
            compositing,
            "Lighting",
            "#F9A825",
            priority=4,
            for_entity="Shot",
        )
        tasks_service.get_or_create_task_type(
            fx, "FX", "#26C6DA", priority=5, for_entity="Shot"
        )
        tasks_service.get_or_create_task_type(
            compositing,
            "Rendering",
            "#F06292",
            priority=6,
            for_entity="Shot",
        )
        tasks_service.get_or_create_task_type(
            compositing,
            "Compositing",
            "#ff5252",
            priority=7,
            for_entity="Shot",
        )
        tasks_service.get_or_create_task_type(
            compositing,
            "Edit",
            "#9b298c",
            priority=8,
            for_entity="Edit",
        )

        tasks_service.get_or_create_task_type(
            concept, "Concept", "#8D6E63", 1, for_entity="Concept"
        )

        print("Task types initialized.")

        tasks_service.get_default_status()
        tasks_service.get_or_create_status(
            "Work In Progress", "wip", "#3273dc", is_wip=True
        )
        tasks_service.get_or_create_status(
            "Waiting For Approval", "wfa", "#ab26ff", is_feedback_request=True
        )
        tasks_service.get_or_create_status(
            "Retake", "retake", "#ff3860", is_retake=True
        )
        tasks_service.get_or_create_status(
            "Done", "done", "#22d160", is_done=True
        )
        tasks_service.get_or_create_status(
            "Ready To Start", "ready", "#fbc02d"
        )

        tasks_service.get_or_create_status(
            "Neutral",
            "neutral",
            "#CCCCCC",
            is_default=True,
            for_concept=True,
            is_artist_allowed=True,
            is_client_allowed=True,
        )

        tasks_service.get_or_create_status(
            "Approved",
            "approved",
            "#66BB6A",
            for_concept=True,
            is_artist_allowed=True,
            is_client_allowed=True,
        )

        tasks_service.get_or_create_status(
            "Rejected",
            "rejected",
            "#E81123",
            for_concept=True,
            is_artist_allowed=True,
            is_client_allowed=True,
        )

        print("Task status initialized.")


def sync_with_ldap_server():
    """
    Connect to a LDAP server, then creates all related accounts.
    """
    LDAP_HOST = app.config["LDAP_HOST"]
    LDAP_PORT = app.config["LDAP_PORT"]
    LDAP_PASSWORD = os.getenv("LDAP_PASSWORD", "password")
    LDAP_BASE_DN = app.config["LDAP_BASE_DN"]
    LDAP_DOMAIN = app.config["LDAP_DOMAIN"]
    LDAP_USER = os.getenv("LDAP_USER", "")
    LDAP_GROUP = app.config["LDAP_GROUP"]
    LDAP_SSL = app.config["LDAP_SSL"]
    EMAIL_DOMAIN = os.getenv("EMAIL_DOMAIN", "studio.local")
    LDAP_EXCLUDED_ACCOUNTS = os.getenv("LDAP_EXCLUDED_ACCOUNTS", "")
    LDAP_IS_AD = app.config["LDAP_IS_AD"]
    LDAP_IS_AD_SIMPLE = app.config["LDAP_IS_AD_SIMPLE"]

    def clean_value(value):
        cleaned_value = str(value)
        if cleaned_value == "[]":
            cleaned_value = ""
        return cleaned_value

    def search_ldap_users(conn, excluded_accounts):
        is_ad = LDAP_IS_AD or LDAP_IS_AD_SIMPLE
        attributes = ["givenName", "sn", "mail", "cn"]
        if is_ad:
            if LDAP_IS_AD_SIMPLE:
                attributes += ["cn"]
            else:
                attributes += ["sAMAccountName"]
            attributes += [
                "thumbnailPhoto",
                "userAccountControl",
                "objectGUID",
            ]
        else:
            attributes += [
                "uid",
                "jpegPhoto",
                "uniqueIdentifier",
                "organizationalStatus",
            ]
        query = "(objectClass=person)"
        if is_ad:
            query = "(&(objectClass=person)(!(objectClass=computer)))"
        group_members = None
        if len(LDAP_GROUP) > 0:
            if is_ad:
                query = f"(&(objectClass=person)(memberOf={LDAP_GROUP}))"
            else:
                conn.search(
                    LDAP_BASE_DN,
                    f"(&(objectClass=groupofUniqueNames)(cn={LDAP_GROUP}))",
                    attributes=["uniqueMember"],
                )
                group_members = conn.entries[0].uniqueMember.values
        conn.search(LDAP_BASE_DN, query, attributes=attributes)
        ldap_users = []
        for entry in conn.entries:
            if LDAP_IS_AD_SIMPLE:
                desktop_login = entry.cn
            elif LDAP_IS_AD:
                desktop_login = entry.sAMAccountName
            else:
                desktop_login = entry.uid
            desktop_login = clean_value(desktop_login)

            if desktop_login not in excluded_accounts and (
                group_members is None or entry.entry_dn in group_members
            ):
                if is_ad:
                    ldap_uid = clean_value(entry.objectGUID)
                elif entry.uniqueIdentifier:
                    ldap_uid = clean_value(entry.uniqueIdentifier)
                else:
                    ldap_uid = None
                thumbnails = (
                    entry.thumbnailPhoto if is_ad else entry.jpegPhoto
                ).raw_values
                if len(thumbnails) > 0 and len(thumbnails[0]) > 0:
                    thumbnail = thumbnails[0]
                else:
                    thumbnail = None

                emails = entry.mail.values
                if len(emails) == 0:
                    emails = ["%s@%s" % (desktop_login, EMAIL_DOMAIN)]
                else:

                    def sort_mails(email):
                        if email == desktop_login:
                            return -2
                        elif EMAIL_DOMAIN in email:
                            return -1
                        else:
                            return 0

                    emails = sorted(emails, key=sort_mails)

                if is_ad:
                    active = bool(entry.userAccountControl.value & 2) is False
                elif entry.organizationalStatus:
                    active = (
                        entry.organizationalStatus.value.lower() == "active"
                    )
                else:
                    active = False

                ldap_users.append(
                    {
                        "first_name": clean_value(entry.givenName or entry.cn),
                        "last_name": clean_value(entry.sn),
                        "email": emails[0].lower(),
                        "emails": emails,
                        "desktop_login": desktop_login,
                        "thumbnail": thumbnail,
                        "active": active,
                        "ldap_uid": ldap_uid,
                    }
                )
        return ldap_users

    def get_ldap_users():
        excluded_accounts = LDAP_EXCLUDED_ACCOUNTS.split(",")
        ldap_server = f"{LDAP_HOST}:{LDAP_PORT}"
        SSL = LDAP_SSL
        if LDAP_IS_AD_SIMPLE:
            user = LDAP_USER
            authentication = SIMPLE
        elif LDAP_IS_AD:
            user = f"{LDAP_DOMAIN}\\{LDAP_USER}"
            authentication = NTLM
        else:
            user = f"uid={LDAP_USER},{LDAP_BASE_DN}"
            authentication = SIMPLE

        server = Server(ldap_server, get_info=ALL, use_ssl=SSL)
        conn = Connection(
            server,
            user=user,
            password=LDAP_PASSWORD,
            authentication=authentication,
            raise_exceptions=True,
            auto_bind=True,
        )

        return search_ldap_users(conn, excluded_accounts)

    def update_person_list_with_ldap_users(users):
        persons_to_update = []
        persons_to_create = []
        for user in sorted(users, key=lambda k: k["active"]):
            person = None
            try:
                person = persons_service.get_person_by_ldap_uid(
                    user["ldap_uid"]
                )
            except PersonNotFoundException:
                try:
                    person = persons_service.get_person_by_desktop_login(
                        user["desktop_login"]
                    )
                except PersonNotFoundException:
                    for mail in user["emails"]:
                        try:
                            person = persons_service.get_person_by_email(mail)
                            break
                        except PersonNotFoundException:
                            pass

            if person is None:
                persons_to_create.append(user)
            else:
                persons_to_update.append((person, user))

        for person in (
            Person.query.filter_by(is_generated_from_ldap=True, active=True)
            .filter(
                not_(Person.id.in_([p[0]["id"] for p in persons_to_update]))
            )
            .all()
        ):
            persons_service.update_person(
                person.id, {"active": False}, bypass_protected_accounts=True
            )
            print(
                "User %s disabled (not found in LDAP)." % person.desktop_login
            )

        for person, user in persons_to_update:
            try:
                if (
                    not person["active"]
                    and user["active"]
                    and persons_service.is_user_limit_reached()
                ):
                    raise IsUserLimitReachedException

                if any(
                    user[key] != person[key]
                    for key in [
                        key
                        for key in user.keys()
                        if key not in ["thumbnail", "emails"]
                    ]
                ):
                    persons_service.update_person(
                        person["id"],
                        {
                            "email": user["email"],
                            "first_name": user["first_name"],
                            "last_name": user["last_name"],
                            "active": user["active"],
                            "is_generated_from_ldap": True,
                            "desktop_login": user["desktop_login"],
                            "ldap_uid": user["ldap_uid"],
                        },
                        bypass_protected_accounts=True,
                    )
                    print(f"User {user['desktop_login']} updated.")
            except IsUserLimitReachedException:
                print(
                    f"User {user['desktop_login']} update failed (limit reached, limit {config.USER_LIMIT})."
                )
            except BaseException:
                print(
                    f"User {user['desktop_login']} update failed (email duplicated?)."
                )

            if user["thumbnail"] is not None:
                save_thumbnail(person, user["thumbnail"])

        person = None
        for user in persons_to_create:
            if user["active"]:
                try:
                    if persons_service.is_user_limit_reached():
                        raise IsUserLimitReachedException
                    person = persons_service.create_person(
                        user["email"],
                        "default".encode("utf-8"),
                        user["first_name"],
                        user["last_name"],
                        desktop_login=user["desktop_login"],
                        is_generated_from_ldap=True,
                        ldap_uid=user["ldap_uid"],
                    )
                    print(f"User {user['desktop_login']} created.")
                except IsUserLimitReachedException:
                    print(
                        f"User {user['desktop_login']} creation failed (limit reached, limit {config.USER_LIMIT})."
                    )
                except BaseException:
                    print(
                        f"User {user['desktop_login']} creation failed (email duplicated?)."
                    )

            if person is not None and user["thumbnail"] is not None:
                save_thumbnail(person, user["thumbnail"])

    def save_thumbnail(person, thumbnail):
        thumbnail_path = "/tmp/ldap_th.jpg"
        with open(thumbnail_path, "wb") as th_file:
            th_file.write(thumbnail)
        thumbnail_png_path = thumbnail_utils.convert_jpg_to_png(thumbnail_path)
        thumbnail_utils.turn_into_thumbnail(
            thumbnail_png_path, size=thumbnail_utils.BIG_SQUARE_SIZE
        )
        file_store.add_picture("thumbnails", person["id"], thumbnail_png_path)
        os.remove(thumbnail_png_path)
        persons_service.update_person(
            person["id"], {"has_avatar": True}, bypass_protected_accounts=True
        )

    ldap_users = get_ldap_users()
    update_person_list_with_ldap_users(ldap_users)


def import_data_from_another_instance(
    source,
    login,
    password,
    project=None,
    with_events=False,
    no_projects=False,
    only_projects=False,
):
    """
    Retrieve and save all the data from another API instance. It doesn't
    change the IDs.
    """
    with app.app_context():
        sync_service.init(source, login, password)
        if not only_projects:
            sync_service.run_main_data_sync(project=project)
        if not no_projects:
            sync_service.run_project_data_sync(project=project)
            sync_service.run_other_sync(
                project=project, with_events=with_events
            )


def run_sync_change_daemon(event_source, source, login, password, logs_dir):
    """
    Listen to event websocket. Each time a change occurs, it retrieves the
    related data and save it in the current instance.
    """
    with app.app_context():
        event_client = sync_service.init_events_listener(
            source, event_source, login, password, logs_dir
        )
        sync_service.add_main_sync_listeners(event_client)
        sync_service.add_project_sync_listeners(event_client)
        sync_service.add_special_sync_listeners(event_client)
        print("Start listening.")
        sync_service.run_listeners(event_client)


def run_sync_file_change_daemon(
    event_source, source, login, password, logs_dir
):
    """
    Listen to event websocket. Each time a change occurs, it retrieves the
    related file and save it in the current instance storage.
    """
    with app.app_context():
        event_client = sync_service.init_events_listener(
            source, event_source, login, password, logs_dir
        )
        sync_service.add_file_listeners(event_client)
        print("Start listening.")
        sync_service.run_listeners(event_client)


def import_last_changes_from_another_instance(
    source, login, password, minutes=0, limit=300
):
    """
    Retrieve and save all the data related to most recent events from another
    API instance. It doesn't change the IDs.
    """
    with app.app_context():
        sync_service.init(source, login, password)
        print("Last events syncing started.")
        sync_service.run_last_events_sync(minutes=minutes, limit=300)
        print("Last events syncing ended.")


def import_last_file_changes_from_another_instance(
    source, login, password, minutes=20, limit=50, force=False
):
    """
    Retrieve and save all the data related most to recent file events
    from another API instance (new previews and thumbnails).
    It doesn't change the IDs.
    """
    with app.app_context():
        sync_service.init(source, login, password)
        print("Last files syncing started.")
        sync_service.run_last_events_files(minutes=minutes, limit=50)
        print("Last files syncing ended.")


def import_files_from_another_instance(
    source,
    login,
    password,
    project=None,
    multithreaded=False,
    number_workers=30,
    number_attemps=3,
    force_resync=False,
):
    """
    Retrieve and save all the data related most recent events from another API
    instance. It doesn't change the IDs.
    """
    with app.app_context():
        sync_service.init(
            source, login, password, multithreaded, number_workers
        )
        return sync_service.download_files_from_another_instance(
            project=project,
            multithreaded=multithreaded,
            number_workers=number_workers,
            number_attemps=number_attemps,
            force_resync=force_resync,
        )


def download_file_from_storage():
    with app.app_context():
        sync_service.download_entity_thumbnails_from_storage()
        sync_service.download_preview_files_from_storage()


def dump_database(store=False):
    now = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
    filename = f"zou-db-backup-{now}.sql.gz"
    if store:
        filename = os.path.join(tempfile.gettempdir(), filename)

    filename = backup_service.generate_db_backup(
        app.config["DATABASE"]["host"],
        app.config["DATABASE"]["port"],
        app.config["DATABASE"]["username"],
        app.config["DATABASE"]["password"],
        app.config["DATABASE"]["database"],
        filename,
    )

    if store:
        backup_service.store_db_backup(os.path.basename(filename), filename)
        os.remove(filename)
        print(
            f"Postgres dump added to store (dbbackup/{os.path.basename(filename)})."
        )
    else:
        print(f"Postgres dump created ({os.path.realpath(filename)}).")


def upload_files_to_cloud_storage(days):
    with app.app_context():
        backup_service.upload_entity_thumbnails_to_storage(days)
        backup_service.upload_preview_files_to_storage(days)


def reset_tasks_data(project_id):
    with app.app_context():
        tasks_service.reset_tasks_data(project_id)


def remove_old_data(days_old=90):
    with app.app_context():
        print("Start removing non critical data older than %s." % days_old)
        print("Removing old events...")
        deletion_service.remove_old_events(days_old)
        print("Removing old login logs...")
        deletion_service.remove_old_login_logs(days_old)
        print("Removing old notitfications...")
        deletion_service.remove_old_notifications(days_old)
        print("Old data removed.")


def reset_search_index():
    with app.app_context():
        print("Resetting search index.")
        index_service.reset_index()
        print("Search index resetted.")


def search_asset(query):
    with app.app_context():
        assets = index_service.search_assets(query)
        if len(assets) == 0:
            print("No asset found")
        for asset in assets:
            print(asset["name"], asset["id"])
        return assets


def generate_preview_extra(
    project=None,
    entity_id=None,
    episodes=[],
    only_shots=False,
    only_assets=False,
    force_regenerate_tiles=False,
    with_tiles=False,
    with_metadata=False,
    with_thumbnails=False,
):
    with app.app_context():
        preview_files_service.generate_preview_extra(
            project=project,
            entity_id=entity_id,
            episodes=episodes,
            only_shots=only_shots,
            only_assets=only_assets,
            force_regenerate_tiles=force_regenerate_tiles,
            with_thumbnails=with_thumbnails,
            with_metadata=with_metadata,
            with_tiles=with_tiles,
        )


def reset_movie_files_metadata():
    with app.app_context():
        preview_files_service.reset_movie_files_metadata()


def reset_picture_files_metadata():
    with app.app_context():
        preview_files_service.reset_picture_files_metadata()


def reset_breakdown_data():
    with app.app_context():
        print("Resetting breakdown data for all open projects.")
        breakdown_service.refresh_all_shot_casting_stats()
        print("Resetting done.")


def create_bot(
    email,
    name,
    expiration_date,
    role,
):
    with app.app_context():
        # Allow "admin@example.com" to be invalid.
        if email != "admin@example.com":
            auth.validate_email(email)
        bot = persons_service.create_person(
            email=email,
            password=None,
            first_name=name,
            last_name="",
            expiration_date=expiration_date,
            role=role,
            is_bot=True,
        )
        print(bot["access_token"])


def renormalize_movie_preview_files(
    preview_file_id=None,
    project_id=None,
    all_broken=None,
    all_processing=None,
    days=None,
    hours=None,
    minutes=None,
):
    with app.app_context():
        if preview_file_id is None and not all_broken and not all_processing:
            print(
                "You must specify at least one flag from --all-broken or --all-processing."
            )
            sys.exit(1)

        query = PreviewFile.query.filter(
            PreviewFile.extension == "mp4"
        ).order_by(PreviewFile.created_at.asc())

        if any((minutes, hours, days)):
            since_date = datetime.datetime.now() - datetime.timedelta(
                days=days or 0,
                hours=hours or 0,
                minutes=minutes or 0,
            )
            query = query.filter(PreviewFile.created_at >= since_date)

        if preview_file_id is not None:
            query = query.filter(PreviewFile.id == preview_file_id)

        if project_id is not None:
            query = query.join(Task).filter(
                PreviewFile.project_id == project_id
            )

        if all_broken and all_processing:
            query = query.filter(
                PreviewFile.status.in_(("broken", "processing"))
            )
        elif all_broken:
            query = query.filter(PreviewFile.status == "broken")
        elif all_processing:
            query = query.filter(PreviewFile.status == "processing")

        preview_files = query.all()
        len_preview_files = len(preview_files)
        if len_preview_files == 0:
            print("No preview files found.")
            sys.exit(1)
        else:
            for i, preview_file in enumerate(preview_files):
                try:
                    preview_file_id = str(preview_file.id)
                    print(
                        f"Renormalizing preview file {preview_file_id} ({i+1}/{len_preview_files})."
                    )
                    extension = preview_file.extension
                    uploaded_movie_path = os.path.join(
                        config.TMP_DIR,
                        f"{preview_file_id}.{extension}.tmp",
                    )
                    try:
                        if config.FS_BACKEND == "local":
                            shutil.copyfile(
                                file_store.get_local_movie_path(
                                    "source", preview_file_id
                                ),
                                uploaded_movie_path,
                            )
                        else:
                            sync_service.download_file(
                                uploaded_movie_path,
                                "source",
                                file_store.open_movie,
                                str(preview_file_id),
                            )
                    except:
                        pass
                    if config.ENABLE_JOB_QUEUE:
                        queue_store.job_queue.enqueue(
                            preview_files_service.prepare_and_store_movie,
                            args=(
                                preview_file_id,
                                uploaded_movie_path,
                                True,
                                False,
                            ),
                            job_timeout=int(config.JOB_QUEUE_TIMEOUT),
                        )
                    else:
                        preview_files_service.prepare_and_store_movie(
                            preview_file_id,
                            uploaded_movie_path,
                            normalize=True,
                            add_source_to_file_store=False,
                        )
                except Exception as e:
                    print(
                        f"Renormalization of preview file {preview_file_id} failed: {e}"
                    )
                    continue


def list_plugins(output_format, verbose, filter_field, filter_value):
    with app.app_context():
        query = Plugin.query

        # Apply filter if needed
        if filter_field and filter_value:
            if filter_field == "maintainer":
                query = query.filter(
                    Plugin.maintainer_name.ilike(f"%{filter_value}%")
                )
            else:
                model_field = getattr(Plugin, filter_field)
                query = query.filter(model_field.ilike(f"%{filter_value}%"))

        plugins = query.order_by(Plugin.name).all()

        if not plugins:
            click.echo("No plugins found matching the criteria.")
            return

        plugin_list = []
        for plugin in plugins:
            maintainer = (
                f"{plugin.maintainer_name} <{plugin.maintainer_email}>"
                if plugin.maintainer_email
                else plugin.maintainer_name
            )
            plugin_data = {
                "Plugin ID": plugin.plugin_id,
                "Name": plugin.name,
                "Version": plugin.version,
                "Maintainer": maintainer,
                "License": plugin.license,
            }
            if verbose:
                plugin_data["Description"] = plugin.description or "-"
                plugin_data["Website"] = plugin.website or "-"
                plugin_data["Revision"] = plugin.revision or "-"
                plugin_data["Installation Date"] = plugin.created_at
                plugin_data["Last Update"] = plugin.updated_at
            plugin_list.append(plugin_data)

        if output_format == "table":
            headers = plugin_list[0].keys()
            rows = [p.values() for p in plugin_list]
            click.echo(tabulate(rows, headers, tablefmt="fancy_grid"))
        elif output_format == "json":
            click.echo(json.dumps(plugin_list, indent=2, ensure_ascii=False))



from flask import current_app

from zou.app.models.task_type import TaskType
from zou.app.models.task_status import TaskStatus
from zou.app.models.project import Project
from zou.app.models.person import Person
from zou.app.models.task import Task

from zou.app.services import deletion_service, tasks_service

from zou.app.blueprints.source.shotgun.base import (
    BaseImportShotgunResource,
    ImportRemoveShotgunBaseResource,
)


class ImportShotgunTasksResource(BaseImportShotgunResource):
    def prepare_import(self):
        self.project_ids = Project.get_id_map()
        self.person_ids = Person.get_id_map()
        self.task_type_ids = TaskType.get_id_map(field="shotgun_id")
        self.task_status_ids = TaskStatus.get_id_map(field="short_name")

    def filtered_entries(self):
        return [x for x in self.sg_entries if self.is_valid_task(x)]

    def is_valid_task(self, sg_task):
        return sg_task["step"] is not None and sg_task["project"] is not None

    def extract_data(self, sg_task):
        entity_id = self.get_entity_id(sg_task["entity"])
        task_status_id = self.task_status_ids.get(
            sg_task["sg_status_list"], None
        )
        step_shotgun_id = sg_task["step"]["id"]
        assigner_id = self.person_ids.get(sg_task["created_by"]["id"], None)
        project_id = self.project_ids.get(sg_task["project"]["id"], None)
        task_type_id = self.task_type_ids.get(step_shotgun_id, None)
        assignees = self.extract_assignees(sg_task, self.person_ids)

        return {
            "name": sg_task["cached_display_name"],
            "shotgun_id": sg_task["id"],
            "start_date": sg_task["start_date"],
            "due_date": sg_task["due_date"],
            "sort_order": sg_task["sg_sort_order"],
            "duration": sg_task["duration"],
            "task_type_id": task_type_id,
            "task_status_id": task_status_id,
            "project_id": project_id,
            "entity_id": entity_id,
            "assigner_id": assigner_id,
            "assignees": assignees,
        }

    def get_entity_id(self, sg_entity):
        entity_id = None
        if sg_entity is not None:
            entity_sg_id = sg_entity["id"]
            entity_type = sg_entity["type"]
            if entity_type == "Asset":
                entity_id = self.get_asset_id(entity_sg_id)
            elif entity_type == "Shot":
                entity_id = self.get_shot_id(entity_sg_id)
            elif entity_type == "Scene":
                entity_id = self.get_scene_id(entity_sg_id)
            elif entity_type == "Sequence":
                entity_id = self.get_sequence_id(entity_sg_id)
            elif entity_type == "Episode":
                entity_id = self.get_episode_id(entity_sg_id)
            else:
                return None

        return entity_id

    def extract_assignees(self, sg_task, person_ids):
        assignees = []
        if len(sg_task["task_assignees"]) > 0:
            for sg_person in sg_task["task_assignees"]:
                person_id = person_ids[sg_person["id"]]
                person = Person.get(person_id)
                assignees.append(person)
        return assignees

    def import_entry(self, data):
        task = Task.get_by(shotgun_id=data["shotgun_id"])

        if task is None:
            task = Task.get_by(
                name=data["name"],
                project_id=data["project_id"],
                task_type_id=data["task_type_id"],
                entity_id=data["entity_id"],
            )

        if task is None:
            task = Task(**data)
            task.save()
            current_app.logger.info("Task created: %s" % task)
        else:
            existing_task = Task.get_by(
                name=data["name"],
                project_id=data["project_id"],
                task_type_id=data["task_type_id"],
                entity_id=data["entity_id"],
            )

            if existing_task is not None:
                data.pop("name", None)
                data.pop("project_id", None)
                data.pop("task_type_id", None)
                data.pop("entity_id", None)

            task.update(data)
            tasks_service.clear_task_cache(str(task.id))
            current_app.logger.info("Task updated: %s" % task)

        return task


class ImportRemoveShotgunTaskResource(ImportRemoveShotgunBaseResource):
    def __init__(self):
        ImportRemoveShotgunBaseResource.__init__(self, Task, self.delete_func)

    def delete_func(self, entity):
        deletion_service.remove_task(entity.id)
        tasks_service.clear_task_cache(str(entity.id))



"""add is_casting_standby column to Entity

Revision ID: b97a71306fc8
Revises: 5a291251823c
Create Date: 2022-07-12 14:21:29.530095

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "b97a71306fc8"
down_revision = "5a291251823c"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "entity", sa.Column("is_casting_standby", sa.Boolean(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("entity", "is_casting_standby")
    # ### end Alembic commands ###



"""add archived field to software

Revision ID: 9683bd840dee
Revises: 4bd9bfb73f11
Create Date: 2025-07-14 12:42:05.362508

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "9683bd840dee"
down_revision = "4bd9bfb73f11"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("software", schema=None) as batch_op:
        batch_op.add_column(sa.Column("archived", sa.Boolean(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("software", schema=None) as batch_op:
        batch_op.drop_column("archived")

    # ### end Alembic commands ###
